/*
  Author: Ruriko Yoshida 
  Code  : This is a C++ version of Asger Hobolth's R code.
  Date  : April 12th., 2006
  Update: April 16th, 2006
  Note  : This cpp file will be eventually attached to EMGNJ.cpp
          file.  (i.e. this will be some sort of library.)

*/

#include "TreeNode.h"
#include "Newick.h"
#include <lapack++.h>
#include <lapack.h>
#include <lafnames.h>
#include <iostream>
#include "Felsenstein.h"
#include "Mjoin.h"
#include "Seq.h"
#include "Timer.h"
#include "NJTree.h"
#include "JC-Tree.h"
#include "cmpressed.h"


using namespace std;


int main(int argc, char **argv){
  /* Precision for printing.... */
  cout.precision(7);
  cout.setf(ios::scientific, ios::floatfield);

  /* Declearing variables.... */
  int numSt = 4, numNds = 0, m = 3, numLvs = 0, ** PhylMat, **SeqMat, nIter = 10, nZell = 50, BignIter = 10;
  TreeNode* initialTree;
  int SeqLen = 0, newLn = 0, h = 0;
  LaVectorDouble BrLenV2;
  long double logL = 0.0, totalLog = 0.0;
  LaSymmMatDouble RateMat(numSt, numSt), Q(numSt, numSt);
  char filename[127], Newick[127], freq[127], TsTv[127];
  Alinements * InputData;
  LaVectorDouble FreqVec(numSt), Tht(Choose(numSt, 2)); 
  double  ThtUpdt[Choose(numSt, 2)], trans = 0.5;
  int **PhlMat, **cmpMatrix;
  ofstream finalTreeFile;
  char ** listOfNames; 
  double tfreq = 0.0, cfreq = 0.0, afreq = 0.0, gfreq = 0.0;
  char invocation[200], mOption[127], mfile[127], iter[127];
  Timer ct;

  /* Reading input files and arguements.... */
  strcpy(mOption, "no");
  strcpy(mfile, "no");
  strcpy(iter, "no");
  strcpy(freq, "no");
  strcpy(TsTv, "no");

  cout << "EMGNJ v0.9" << endl;
  m = 2;
  strcpy(filename, argv[argc-1]);

  strcpy(invocation, "invocation: ");
  for(int f = 0; f < argc; f++){
    strcat(invocation, argv[f]);
    strcat(invocation, " ");
    if(strncmp(argv[f], "-m", 2) == 0) 
      { strcpy(mOption, "yes"); 
      if((argv[f + 1] == NULL) || (atoi(argv[f + 1]) < 2) || (atof(argv[f + 1]) != atoi(argv[f+1]))){
	cerr << "Error: Please make sure that m is an integer > 1." << endl;
	exit(0);
      }
      else m = atoi(argv[f + 1]);
      }
    if(strncmp(argv[f], "-f", 2) == 0) {

      /* Checking freq for T */
      strcpy(freq, "yes");
      if((argv[f + 1] == NULL) || (atof(argv[f + 1]) < 0)){
	cerr << "Must enter an integer for frequency." << endl;
	exit(0);
      }
      else  tfreq = atof(argv[f + 1]);

      /* Checking freq for C */
      
      if((argv[f + 2] == NULL) || (atof(argv[f + 2]) < 0)){
	cerr << "Must enter an integer for frequency." << endl;
	exit(0);
      }
      else  cfreq = atof(argv[f + 2]);

      /* Checking freq for A */
      
      if((argv[f + 3] == NULL) || (atof(argv[f + 3]) < 0)){
	cerr << "Must enter an integer for frequency." << endl;
	exit(0);
      }
      else  afreq = atof(argv[f + 3]);

      /* Checking freq for G */
      
      if((argv[f + 4] == NULL) || (atof(argv[f + 4]) < 0)){
	cerr << "Must enter an integer for frequency." << endl;
	exit(0);
      }
      else  gfreq = atof(argv[f + 4]);
    }

    if(strncmp(argv[f], "-t", 2) == 0) {
      strcpy(TsTv, "yes");
      if((argv[f + 1] == NULL) || (atof(argv[f + 1]) < 0)){
	cerr << "Must enter an integer for transition bias." << endl;
	exit(0);
      }
      else  trans = atof(argv[f + 1]);
    }

    if(strncmp(argv[f], "-r", 2) == 0)    strcpy(mfile, "yes");
    if(strncmp(argv[f], "-I", 2) == 0)
      { strcpy(iter, "yes"); 
      if((argv[f + 1] == NULL) || (atoi(argv[f + 1]) < 2) || (atof(argv[f + 1]) != atoi(argv[f+1]))){
	cerr << "Error: Please make sure that m is an integer > 1." << endl;
	exit(0);
      }
      else nIter = atoi(argv[f + 1]);
      }
  }

  cout << invocation << endl;
 
  if(argc < 2){
    cerr << "Please enter an input file name!" << endl;
    exit(0);
  }
//   nIter = atoi(argv[1]);
//   strcpy(filename, argv[argc - 1]);
  strcpy(Newick, "no");
  
  
  /* Setting the GTR model..... */
  if(freq[0] == 'n')
    {
      FreqVec(0) = 0.25; FreqVec(1) = 0.25;  FreqVec(2) = 0.25;  FreqVec(3) = 0.25;
    }
  else{ 
    FreqVec(0) = afreq; FreqVec(1) = gfreq;  FreqVec(2) = cfreq;  FreqVec(3) = tfreq; }

  if(freq[0] == 'y') 
    cout << fixed << "Frequencies: T:" << FreqVec(3) << " C:" << FreqVec(2) << " A:" << FreqVec(0) << " G:" << FreqVec(1) << endl;

  if(TsTv[0] == 'y')
    cout << "Transition/transversion ratio is: " << trans << endl;

  Tht(0) = 1;  Tht(1) = 1; Tht(2) = 1;  Tht(3) = 1;  Tht(4) = 1; Tht(5) = 1;
//   RateMat = GTRRateMFct(FreqVec, Tht, numSt);
//   cout << fixed << RateMat << endl;
  

//   strcpy(freq, "yes");
  strcpy(TsTv, "yes");

  /* Initializing variables...... */  
  PhylMat = new int*[numLvs + numNds];
  for(int i = 0; i < (numLvs + numNds); i++)
    PhylMat[i] = new int[4];
  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < 4; j++) 
      PhylMat[i][j] = 0;
  
  numLvs = 0;
  ct.reset();

  /* Writing tree matrix from DNA sequences using MJOIN to get initial Branchlengths..... */
  PhylMat = writeTreeMatrix(filename, Newick, numLvs, m, BrLenV2, &InputData,
			    FreqVec(3), FreqVec(2),FreqVec(0), FreqVec(1),  
			    freq, TsTv, trans, &listOfNames, &SeqMat, SeqLen, 
			    numSt);
  numNds = numLvs - 1; 
  int dim = (2 * numLvs) - 1; 
  
//   for(int i = 0; i < numLvs; i++) cout << listOfNames[i] << endl;
//   exit(0);

//     for(int i = 0; i < dim; i++){
//       for(int j = 0; j < 4; j++)
//         cout << PhylMat[i][j] << " ";
//       cout << endl;
//   }
  
// /* just for testing.... */
//   for(int i = 0; i < (numLvs + numNds - 1); i++)
//     BrLenV2(i) = 0.01;

  /* Define variables .... */
  LaGenMatDouble CndProbMat(numNds + numLvs - 1, numSt * numSt),
    totalCndProbMat(numNds + numLvs - 1, numSt * numSt);
  //SuffStatMat((numNds + numLvs - 1), numSt * numSt), 
  double FMat[(numNds + numLvs - 1)][Choose(numSt, 2)], 
    GMat[(numNds + numLvs - 1)][Choose(numSt, 2)];
  LaGenMatDouble InvFEgnVc(numSt, numSt), TspFEgnVc(numSt, numSt), 
    EgnVc(numSt, numSt), tmpMat(numSt * numSt, numSt * numSt);
  LaVectorDouble EgnVl(numSt);

  long int numOfSubsets = Choose(numLvs, m);
  int **subsets;
  long double * totalBrLen;

  subsets = new int*[numOfSubsets];
  for(int i = 0; i < numOfSubsets; i++) subsets[i] = new int[m];
  getList3(numLvs, m, subsets);
  for(int i = 0; i < numOfSubsets; i++)
    for(int j = 0; j < m; j++) subsets[i][j]++;

  totalBrLen = new long double[numOfSubsets];

  int tmpVec[numLvs], nObs = 0;
  double Fsum = 0.0, Gsum = 0.0;
  double SuffStatMat[(numNds + numLvs - 1)][numSt * numSt];
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      totalCndProbMat(i, j) = 0.0;
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      SuffStatMat[i][j] = 0.0;
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < Choose(numSt, 2); j++)
      FMat[i][j] = 0.0;
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < Choose(numSt, 2); j++)
      GMat[i][j] = 0.0;
 
//   SeqMat = DNAMat(InputData, numLvs, SeqLen);
  cmpMatrix = cmpressed(SeqMat, numLvs, newLn, SeqLen);
//   for(int i = 0; i < (numLvs + 1); i++){
//     for(int j = 0; j < newLn; j++)
//       cout << cmpMatrix[i][j] << " ";
//     cout << endl;
//   }

//  getSubtreeWt(PhylMat, numLvs, 4, leaves, BrLenV2);  exit(0);
  /* Get emp freq.... */
//   getEmpFreq(FreqVec, SeqMat,  SeqLen, numLvs, numSt);

  /* Printing out initial information.... */
  cout << "Estimated Emp Frq: "; 
  for(int i = 0; i < numSt; i++)
  cout << fixed << FreqVec(i) << " ";
  cout << endl;
  cout << "Initial Theta: ";
  for(int i = 0; i < Choose(numSt, 2); i++)
    cout << fixed << Tht(i) << " ";
  cout << endl;
  cout << "Initial Branch lengths: ";
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    cout << fixed << BrLenV2(i) << " ";
  cout << endl;
  cout << endl;

  /* Big iteration loop... */


  for(int big = 0; big < BignIter; big++){
    cout << "Tree Iteration: " << big + 1 << endl;
    for(int iter = 0; iter < nIter; iter++){
      cout << "Iteration: " << iter + 1 << endl;
      for(int i = 0; i < (numLvs + numNds - 1); i++)
	for(int j = 0; j < (numSt * numSt); j++)
	  totalCndProbMat(i, j) = 0.0;
      
      totalLog = 0.0;
      
      for(int i = 0; i < (numLvs + numNds - 1); i++)
	for(int j = 0; j < (numSt * numSt); j++)
	  SuffStatMat[i][j] = 0.0;
      for(int i = 0; i < (numLvs + numNds - 1); i++)
	for(int j = 0; j < Choose(numSt, 2); j++)
	  FMat[i][j] = 0.0;
      for(int i = 0; i < (numLvs + numNds - 1); i++)
	for(int j = 0; j < Choose(numSt, 2); j++)
	  GMat[i][j] = 0.0;
      
      RateMat = GTRRateMFct(FreqVec, Tht, numSt);
      
      computeEvalEvec(RateMat, InvFEgnVc, TspFEgnVc, EgnVl, EgnVc, FreqVec, numSt);
      //     cout << InvFEgnVc << endl << endl;
      //     cout << TspFEgnVc << endl << endl;
      
      for(int site = 0; site < newLn; site++){
	for(int j = 0; j < numLvs; j++) tmpVec[j] = 0;
	
	for(int j = 0; j < numLvs; j++) tmpVec[j] = cmpMatrix[j][site];
	//       for(int j = 0; j < numLvs; j++) cout << tmpVec[j] << " ";
	//       cout << endl; exit(0);
	//       FlsnFct(PhylMat, RateMat, FreqVec, BrLenV2, tmpVec, numSt, numLvs,
	// 	      numNds, CndProbMat, logL);
	
	FstFlsnFct(PhylMat, EgnVl, InvFEgnVc, TspFEgnVc, BrLenV2, FreqVec, 
		   tmpVec, numSt, numLvs, numNds, CndProbMat, logL);
	
	nObs = cmpMatrix[numLvs][site];
	
	for(int i = 0; i < (numLvs + numNds - 1); i++)
	  for(int j = 0; j < (numSt * numSt); j++)
	    totalCndProbMat(i, j) += ((double) nObs) * CndProbMat(i, j);
	totalLog += ((double) nObs) * logL;
      }
      
      cout << "logL: " << totalLog << endl;
      //    cout << fixed << CndProbMat << endl; exit(0);
      
      for(int j = 0; j < (numLvs + numNds-1); j++){
	
	//      for(int k = 0; k < (numSt * numSt); k++) hold[k] = 0.0;
	for(int i = 0; i < numSt; i++)
	  for(int k = 0; k < numSt; k++) Q(i, k) = RateMat(i, k) * BrLenV2(j);
	
	for(int i = 0; i < (numSt * numSt); i++)
	  for(int k = 0; k < (numSt * numSt); k++) tmpMat(i, k) = 0.0;
	tmpMat = Transposition(CndStatFct(FreqVec, Q, numSt), numSt*numSt);
	
	for(int ii = 0; ii < (numSt * numSt); ii++)
	  for(int jj = 0; jj < (numSt * numSt); jj++)
	    // 	  SuffStatMat[j][ii] += totalCndProbMat(j, jj) * (Transposition(CndStatFct(FreqVec, Q, numSt), numSt*numSt))(jj, ii);  
	    SuffStatMat[j][ii] += totalCndProbMat(j, jj) * tmpMat(jj, ii);  
	
	//      cout << fixed << Q << endl<< BrLenV2(j) <<endl; 
	//     cout << fixed << CndStatFct(FreqVec, Q, numSt) << endl<< endl; 
	//     cout << fixed << Transposition(CndStatFct(FreqVec, Q, numSt),  numSt * numSt) << endl<< endl;
	
	//       for(int jj = 0; jj < (numSt * numSt); jj++)
	// 	SuffStatMat(j, jj) = hold[jj];
      }
      
      //     cout << fixed << SuffStatMat << endl<< endl;
      //     cout << fixed << CndStatFct(FreqVec, Q, numSt) << endl<< endl; 
      //     cout << fixed << Transposition(CndStatFct(FreqVec, Q, numSt),  numSt * numSt) << endl<< endl; exit(0);
      
      
      for(int j = 0; j < (numLvs + numNds-1); j++){  
	FMat[j][0] = SuffStatMat[j][4] + SuffStatMat[j][7];
	FMat[j][1] = SuffStatMat[j][5] + SuffStatMat[j][10];
	FMat[j][2] = SuffStatMat[j][6] + SuffStatMat[j][13];
	FMat[j][3] = SuffStatMat[j][8] + SuffStatMat[j][11];
	FMat[j][4] = SuffStatMat[j][9] + SuffStatMat[j][14];
	FMat[j][5] = SuffStatMat[j][12] + SuffStatMat[j][15];
      }
      for(int j = 0; j < (numLvs + numNds - 1); j++){ 
	GMat[j][0] = (SuffStatMat[j][0] * FreqVec(1)) + (SuffStatMat[j][1] * FreqVec(0));
	GMat[j][1] = (SuffStatMat[j][0] * FreqVec(2)) + (SuffStatMat[j][2] * FreqVec(0));
	GMat[j][2] = (SuffStatMat[j][0] * FreqVec(3)) + (SuffStatMat[j][3] * FreqVec(0));
	GMat[j][3] = (SuffStatMat[j][1] * FreqVec(2)) + (SuffStatMat[j][2] * FreqVec(1));
	GMat[j][4] = (SuffStatMat[j][1] * FreqVec(3)) + (SuffStatMat[j][3] * FreqVec(1));
	GMat[j][5] = (SuffStatMat[j][2] * FreqVec(3)) + (SuffStatMat[j][3] * FreqVec(2));
      }
      
      //    cout << fixed << FMat << endl;
      //     cout << fixed << GMat << endl;
      for(int j = 0; j < Choose(numSt, 2); j++) ThtUpdt[j] = 0.0;
      
      for(int zell = 0; zell < nZell; zell++){
	for(int j = 0; j < Choose(numSt, 2); j++){
	  Gsum = 0.0; Fsum = 0.0;
	  for(int k = 0; k < (numLvs + numNds - 1); k++){
	    Fsum += FMat[k][j];  Gsum += (GMat[k][j] * BrLenV2(k));
	  }
	  ThtUpdt[j] = Fsum / Gsum;
	  //cout << Gsum << endl;
	}
	
	for(int jj = 0; jj < (numLvs + numNds - 1); jj++){
	  Gsum = 0.0; Fsum = 0.0;
	  for(int k = 0; k < Choose(numSt, 2); k++){
	    Fsum += FMat[jj][k];  Gsum += (GMat[jj][k] * Tht(k));
	  }
	  BrLenV2(jj) = Fsum / Gsum;//cout << Gsum << endl;
	  if(BrLenV2(jj) < 0.0000001) BrLenV2(jj) = 0.0000001;
	}
	for(int k = 0; k < Choose(numSt, 2); k++){
	  if(ThtUpdt[k] < 0.0000001) Tht(k) = 0.0000001;
	  else
	    Tht(k) = ThtUpdt[k];
	}
	Tht(1) = 1.0;
	
	//     for(int k = 0; k < Choose(numSt, 2); k++)
	//       Tht(k) = ThtUpdt[k];
      }
      
      cout << "BrLen: "; 
      for(int j = 0; j < (numLvs + numNds - 1); j++)
      cout << fixed << BrLenV2(j) << " ";
      cout << endl << "Tht: ";
      for(int k = 0; k < Choose(numSt, 2); k++)
	cout << fixed << Tht(k) << " ";
      cout << endl << endl;
      
    }/* End of Big iteration loop... */

    /* Computing subtree weights. */
    cout << "Computing subtree weights....";
    for(int i = 0; i < numOfSubsets; i++){
      totalBrLen[i] = 0.0;
      //     if((i % 50) == 0)
      //       cout << "Weights: " << i + 1 << "/" <<  numOfSubsets << endl;
      totalBrLen[i] = getSubtreeWt(PhylMat, numLvs, m, subsets[i], BrLenV2);
    }cout << "end." << endl;
    
    /* Constructing a tree with the total subBrLen. */
    initialTree = MJOIN(numLvs, m, numOfSubsets, listOfNames, totalBrLen);
    
    finalTreeFile.open("final.dnd");
    printDND2(initialTree, finalTreeFile);
    cout << "The final tree is: " << endl;
    printDND3(initialTree);
    
    h = 0;
    assignLeafLevel(initialTree, h);
    assignNum(initialTree, h, numLvs);
    
    TreeMatrix(PhylMat, initialTree, dim, BrLenV2);
    //     for(int i = 0; i < dim; i++){
    //       for(int j = 0; j < 4; j++)
    //         cout << PhylMat[i][j] << " ";
    //       cout << endl;
    //   }
    deleteTree(initialTree);
  }/* End of Big tree iteration loop... */
  
  
  for(int i = 0; i < numOfSubsets; i++) delete [] subsets[i];
  delete [] subsets;
  delete [] totalBrLen;
  return 0;
}


