/*
  Author: Ruriko Yoshida 
  Code  : This is a C++ version of Asger Hobolth's R code.
  Date  : Dec 8th., 2005
  Update: Dec 9th., 2005

*/

#include <lapack++.h>
#include <lapack.h>
#include <lafnames.h>
#include <iostream>

using namespace std;

void Transposition(LaGenMatDouble A, LaGenMatDouble & B, int N){
  for(int i = 0; i < N; i++)
    for(int j = 0; j < N; j++)
      B(i, j) = A(j, i);
}

LaGenMatDouble Transposition(LaGenMatDouble A, int N){
  LaGenMatDouble B(N, N);
  for(int i = 0; i < N; i++)
    for(int j = 0; j < N; j++)
      B(i, j) = A(j, i);
  return B;
}

void Mult(LaVectorDouble v, LaSymmMatDouble A, LaGenMatDouble & B, int N){
  for(int i = 0; i < N; i++){
    for(int j = 0; j < N; j++){
      B(i, j) = v(i) * A(i, j);
    }
  }
}

void Mult(LaVectorDouble v, LaGenMatDouble A, LaGenMatDouble & B, int N){
  for(int i = 0; i < N; i++){
    for(int j = 0; j < N; j++){
      B(i, j) = v(i) * A(i, j);
    }
  }
}


LaGenMatDouble RateToProbFun(LaSymmMatDouble RateMat, 
		   double Tau, LaVectorDouble FreqVec, int StateNum){
  LaGenMatDouble ProbMat(StateNum, StateNum);
  LaGenMatDouble Diag(StateNum, StateNum), EXP(StateNum, StateNum);
  LaGenMatDouble EigenRate(StateNum, StateNum), T(StateNum, StateNum),  B(StateNum, StateNum);
  LaVectorDouble EVal(StateNum), Freq2(StateNum);
  
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++){
      ProbMat(i, j) = 0.0; Diag(i, j) = 0.0; EXP(i, j) = 0.0; 
      EigenRate(i, j) = 0.0; T(i, j) = 0.0; B(i, j) = 0.0;
    }
  for(int i = 0; i < StateNum; i++){
    EVal(i) = 0.0; Freq2(i) = 0.0;
  }
  
  for(int i = 0; i < StateNum; i++){
    FreqVec(i) = sqrt(FreqVec(i)); Diag(i, i) = 1/FreqVec(i);
    Freq2(i) = 1/FreqVec(i); }
  
  Mult(FreqVec, RateMat, B, StateNum);
  
  B = B * Diag;
  
  LaSymmMatDouble A(StateNum,StateNum);
  LaGenerateMatDouble(A);
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++)
      A(i,j) = B(i,j);
  
  LaEigSolve(A, EVal, EigenRate);  
  for(int i = 0; i < StateNum; i++) EXP(i, i) = exp(Tau * EVal(i));
  
  Mult(FreqVec, EigenRate, B, StateNum); 
  Transposition(B, T, StateNum);
  Mult(Freq2, EigenRate, B, StateNum); 
  ProbMat = (B * (EXP * T));
  //cout << ProbMat << endl;
  // cout << EVal << endl << EigenRate << endl << A << endl;
  return ProbMat;
}

LaGenMatDouble RateToProbFun2(LaGenMatDouble RateMat, 
		   double Tau, LaVectorDouble FreqVec, int StateNum){
  LaGenMatDouble ProbMat(StateNum, StateNum);
  LaGenMatDouble Diag(StateNum, StateNum), EXP(StateNum, StateNum);
  LaGenMatDouble EigenRate(StateNum, StateNum), T(StateNum, StateNum),  B(StateNum, StateNum);
  LaVectorDouble EVal(StateNum), Freq2(StateNum);
  
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++){
      ProbMat(i, j) = 0.0; Diag(i, j) = 0.0; EXP(i, j) = 0.0; 
      EigenRate(i, j) = 0.0; T(i, j) = 0.0; B(i, j) = 0.0;
    }
  for(int i = 0; i < StateNum; i++){
    EVal(i) = 0.0; Freq2(i) = 0.0;
  }
  
  for(int i = 0; i < StateNum; i++){
    FreqVec(i) = sqrt(FreqVec(i)); Diag(i, i) = 1/FreqVec(i);
    Freq2(i) = 1/FreqVec(i); }
  
  Mult(FreqVec, RateMat, B, StateNum);
  
  B = B * Diag;
  
  //  cout << B << endl;
  LaSymmMatDouble A(StateNum,StateNum);
  LaGenerateMatDouble(A);
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++)
      A(i,j) = B(i,j);
  
  LaEigSolve(A, EVal, EigenRate);  
  for(int i = 0; i < StateNum; i++) EXP(i, i) = exp(Tau * EVal(i));
  
  Mult(FreqVec, EigenRate, B, StateNum); 
  Transposition(B, T, StateNum);
  Mult(Freq2, EigenRate, B, StateNum); 
  ProbMat = (B * (EXP * T));
  //cout << ProbMat << endl;
  // cout << EVal << endl << EigenRate << endl << A << endl;
  return ProbMat;
}

LaSymmMatDouble GTRRateMFct(LaVectorDouble Frq, LaVectorDouble Tht,
			    int numOfSts){
  LaSymmMatDouble RateM(numOfSts, numOfSts);
  LaVectorDouble Tht2 = Tht;
  int counter = 0;
  double tmpRate[numOfSts][numOfSts];
  
  Tht2(2) = Tht(3);
  Tht2(3) = Tht(2);

  for(int i = 0;i < numOfSts; i++){
    for(int j = 0; j < numOfSts; j++){
      tmpRate[i][j] = Frq(j); 
      RateM(i, j) = 0.0;
    }
  }

//   cout << Tht2 << endl;
//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     cout << endl;
//   }

//   cout << endl;

  for(int j = 0;j < numOfSts; j++)
    for(int i = 0; i < j; i++){
      //cout << Tht2(counter) << endl;
      tmpRate[i][j] *= Tht2(counter); counter++;
    }

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     cout << endl;
//   }
//   cout << endl;

  counter = 0;
  for(int j = 0;j < numOfSts; j++)
    for(int i = j + 1; i < numOfSts; i++){
      // cout << "(" << i << " " << j << ") " << endl;
      tmpRate[i][j] *= Tht(counter); counter++;
    }

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     //cout << "(" << i << " " << j << ") ";
//     cout << endl;
//   }
//   cout << endl;

  for(int i = 0;i < numOfSts; i++) 
    tmpRate[i][i] = 0.0;

  for(int i = 0;i < numOfSts; i++)
    for(int j = 0; j < numOfSts; j++)
      if(j != i)
	tmpRate[i][i] -= tmpRate[i][j];

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     //cout << "(" << i << " " << j << ") ";
//     cout << endl;
//   }

//   cout << endl;

  for(int i = 0;i < numOfSts; i++)
    for(int j = 0; j < numOfSts; j++)
      RateM(i, j) = tmpRate[i][j];
  
  //  cout << RateM << endl;

  return(RateM);
}



LaGenMatDouble GTRRateMFct2(LaVectorDouble Frq, LaVectorDouble Tht,
			    int numOfSts){
  LaGenMatDouble RateM(numOfSts, numOfSts);
  LaVectorDouble Tht2 = Tht;
  int counter = 0;
  //  double tmpRate[numOfSts][numOfSts];
  
  Tht2(2) = Tht(3);
  Tht2(3) = Tht(2);
  for(int i = 0;i < numOfSts; i++)
    for(int j = 0; j < numOfSts; j++)
      RateM(i, j) = 0.0;
  
  for(int i = 0;i < numOfSts; i++){
    for(int j = 0; j < numOfSts; j++){
      RateM(i, j) = Frq(j); 
    }
  }

//   cout << Tht2 << endl;
//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << RateM(i, j) << " ";
//     cout << endl;
//   }

//   cout << endl;

  for(int j = 0;j < numOfSts; j++)
    for(int i = 0; i < j; i++){
      //cout << Tht2(counter) << endl;
      RateM(i, j) *= Tht2(counter); counter++;
    }

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     cout << endl;
//   }
//   cout << endl;

  counter = 0;
  for(int j = 0;j < numOfSts; j++)
    for(int i = j + 1; i < numOfSts; i++){
      // cout << "(" << i << " " << j << ") " << endl;
      RateM(i, j) *= Tht(counter); counter++;
    }

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     //cout << "(" << i << " " << j << ") ";
//     cout << endl;
//   }
//   cout << endl;

  for(int i = 0;i < numOfSts; i++) 
    RateM(i, i) = 0.0;

  for(int i = 0;i < numOfSts; i++)
    for(int j = 0; j < numOfSts; j++)
      if(j != i)
	RateM(i, i) -= RateM(i, j);

//   for(int i = 0;i < numOfSts; i++){
//     for(int j = 0; j < numOfSts; j++)
//       cout << tmpRate[i][j] << " ";
//     //cout << "(" << i << " " << j << ") ";
//     cout << endl;
//   }

//   cout << endl;

//   for(int i = 0;i < numOfSts; i++)
//     for(int j = 0; j < numOfSts; j++)
//       RateM(i, j) = tmpRate[i][j];
  
  //  cout << RateM << endl;

  return(RateM);
}



void FlsnFct(int** TreeMatrix, LaSymmMatDouble RateMat, LaVectorDouble Frq, 
	     LaVectorDouble BranchLength, int* Obs, int numSt, int numLvs, 
	     int numNds, LaGenMatDouble & CndProbMat, long double & logL){
  int LftChld = 0, RgtChld = 0, rootLftChld = 0, rootRgtChld = 0, Prnt = 0,
    GrndPrnt = 0, OthChld = 0, count = 0;
  long double LftBrLen = 0.0, RgtBrLen = 0.0, GPBrLen = 0.0, 
    OthBrLen = 0.0;
  LaGenMatDouble UMat(numLvs + numNds, numSt), 
    DMat(numLvs + numNds - 1, numSt);
  // CndProbMat(numNds + numLvs - 1, numSt * numSt); 
  LaVectorDouble tmpVec(numSt), tmpVec2(numSt), tmpVec3(numSt), tmpVec4(numSt);

  /* Initialize.... */
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      CndProbMat(i, j) = 0.0;

  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < numSt; j++){
      UMat(i, j) = 0.0;
    }
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < numSt; j++){
      DMat(i, j) = 0.0;
    }
  for(int i = 0; i < numSt; i++) {
    tmpVec(i) = 0.0; tmpVec2(i) = 0.0; tmpVec3(i) = 0.0; tmpVec4(i) = 0.0; 
  }
  logL = 0.0;

  /* U matrix.... */
  for(int i = 0; i < numLvs; i++)
    UMat(i, Obs[i] - 1) = 1;
  for(int i = 0; i < numNds; i++){
    LftChld = TreeMatrix[numLvs + i][1];
    RgtChld = TreeMatrix[numLvs + i][2];
    LftBrLen = BranchLength(LftChld - 1); RgtBrLen = BranchLength(RgtChld - 1);

    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
    }
  
    for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(LftChld - 1, j);
    tmpVec2 = RateToProbFun(RateMat, LftBrLen, Frq, numSt) * tmpVec;

    for(int j = 0; j < numSt; j++) tmpVec3(j) = UMat(RgtChld - 1, j);
    tmpVec4 = RateToProbFun(RateMat, RgtBrLen, Frq, numSt) * tmpVec3;

    for(int j = 0; j < numSt; j++)
      UMat(numLvs + i, j) = tmpVec2(j)*tmpVec4(j);
  }

  // cout << UMat << endl;
  /* log L */
  for(int j = 0; j < numSt; j++)
    logL += UMat(numNds + numLvs - 1, j) * Frq(j);

  /* D matrix..... */

  LftChld = TreeMatrix[numLvs + numNds - 1][1]; 
  LftBrLen = BranchLength(LftChld - 1);
  RgtChld = TreeMatrix[numLvs + numNds - 1][2]; 
  RgtBrLen = BranchLength(RgtChld - 1);

  rootLftChld = LftChld; rootRgtChld = RgtChld;

  // cout << LftChld << " " << RgtChld << endl;
  for(int j = 0; j < numSt; j++){
    tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
  }

  for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(LftChld - 1, j);
  tmpVec2 = RateToProbFun(RateMat, LftBrLen, Frq, numSt) * tmpVec;
  
  for(int j = 0; j < numSt; j++)
    DMat(RgtChld - 1, j) = Frq(j) * tmpVec2(j);

  for(int j = 0; j < numSt; j++) tmpVec3(j) = UMat(RgtChld - 1, j);
  tmpVec4 = RateToProbFun(RateMat, RgtBrLen, Frq, numSt) * tmpVec3;

  for(int j = 0; j < numSt; j++)
    DMat(LftChld - 1, j) = Frq(j) * tmpVec4(j);
  // cout << DMat << endl;
  for(int i = (numNds + numLvs - 2); i >= 0; i--)
  //  for(int i = 0; i < (numNds + numLvs - 1); i++)
    if(((i + 1) != rootLftChld) && ((i + 1) != rootRgtChld)){    
      Prnt = TreeMatrix[i][3]; 
      GrndPrnt = TreeMatrix[Prnt - 1][3]; GPBrLen = BranchLength(Prnt - 1);
      LftChld = TreeMatrix[Prnt - 1][1]; RgtChld = TreeMatrix[Prnt - 1][2];
      if(LftChld == (i + 1)) OthChld = RgtChld; else OthChld = LftChld;
      OthBrLen = BranchLength(OthChld - 1);
      
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
     }
 
    for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(OthChld - 1, j);
    tmpVec2 = RateToProbFun(RateMat, OthBrLen, Frq, numSt) * tmpVec;
    
    for(int j = 0; j < numSt; j++) tmpVec3(j) = DMat(Prnt - 1, j);

    for(int j = 0; j < numSt; j++)
      for(int k = 0; k < numSt; k++)
	tmpVec4(j) += tmpVec3(k) * (RateToProbFun(RateMat, GPBrLen, Frq, numSt))(k, j);

    for(int j = 0; j < numSt; j++)
      DMat(i, j) = tmpVec2(j)*tmpVec4(j);    
    }
  // cout << DMat << endl;

  /* Computing CndProbMat.... */
  for(int i = 0; i < (numNds + numLvs - 1); i++){
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
    }
    
    for(int j = 0; j < numSt; j++) 
      {tmpVec(j) = DMat(i, j); tmpVec2(j) = UMat(i, j);}
    
    count = 0;
    
    for(int j = 0; j < numSt; j++){
      for(int k = 0; k < numSt; k++){
	CndProbMat(i, count) = ((RateToProbFun(RateMat, BranchLength(i), Frq, numSt))(k, j) * tmpVec(k) * tmpVec2(j))/logL;
	count++;
      }
    }
  }
  logL = log(logL);
}

void FlsnFct2(int** TreeMatrix, LaGenMatDouble RateMat, LaVectorDouble Frq, 
	     LaVectorDouble BranchLength, int* Obs, int numSt, int numLvs, 
	     int numNds, LaGenMatDouble & CndProbMat, long double & logL){
  int LftChld = 0, RgtChld = 0, rootLftChld = 0, rootRgtChld = 0, Prnt = 0,
    GrndPrnt = 0, OthChld = 0, count = 0;
  long double LftBrLen = 0.0, RgtBrLen = 0.0, GPBrLen = 0.0, 
    OthBrLen = 0.0;
  LaGenMatDouble UMat(numLvs + numNds, numSt), 
    DMat(numLvs + numNds - 1, numSt);
  // CndProbMat(numNds + numLvs - 1, numSt * numSt); 
  LaVectorDouble tmpVec(numSt), tmpVec2(numSt), tmpVec3(numSt), tmpVec4(numSt);

  /* Initialize.... */
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      CndProbMat(i, j) = 0.0;

  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < numSt; j++){
      UMat(i, j) = 0.0;
    }
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < numSt; j++){
      DMat(i, j) = 0.0;
    }
  for(int i = 0; i < numSt; i++) {
    tmpVec(i) = 0.0; tmpVec2(i) = 0.0; tmpVec3(i) = 0.0; tmpVec4(i) = 0.0; 
  }
  logL = 0.0;

  /* U matrix.... */
  for(int i = 0; i < numLvs; i++)
    UMat(i, Obs[i] - 1) = 1;
  for(int i = 0; i < numNds; i++){
    LftChld = TreeMatrix[numLvs + i][1];
    RgtChld = TreeMatrix[numLvs + i][2];
    LftBrLen = BranchLength(LftChld - 1); RgtBrLen = BranchLength(RgtChld - 1);

    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
    }
  
    for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(LftChld - 1, j);
    tmpVec2 = RateToProbFun2(RateMat, LftBrLen, Frq, numSt) * tmpVec;

    for(int j = 0; j < numSt; j++) tmpVec3(j) = UMat(RgtChld - 1, j);
    tmpVec4 = RateToProbFun2(RateMat, RgtBrLen, Frq, numSt) * tmpVec3;

    for(int j = 0; j < numSt; j++)
      UMat(numLvs + i, j) = tmpVec2(j)*tmpVec4(j);
  }

  // cout << UMat << endl;
  /* log L */
  for(int j = 0; j < numSt; j++)
    logL += UMat(numNds + numLvs - 1, j) * Frq(j);

  /* D matrix..... */

  LftChld = TreeMatrix[numLvs + numNds - 1][1]; 
  LftBrLen = BranchLength(LftChld - 1);
  RgtChld = TreeMatrix[numLvs + numNds - 1][2]; 
  RgtBrLen = BranchLength(RgtChld - 1);

  rootLftChld = LftChld; rootRgtChld = RgtChld;

  // cout << LftChld << " " << RgtChld << endl;
  for(int j = 0; j < numSt; j++){
    tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
  }

  for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(LftChld - 1, j);
  tmpVec2 = RateToProbFun2(RateMat, LftBrLen, Frq, numSt) * tmpVec;
  
  for(int j = 0; j < numSt; j++)
    DMat(RgtChld - 1, j) = Frq(j) * tmpVec2(j);

  for(int j = 0; j < numSt; j++) tmpVec3(j) = UMat(RgtChld - 1, j);
  tmpVec4 = RateToProbFun2(RateMat, RgtBrLen, Frq, numSt) * tmpVec3;

  for(int j = 0; j < numSt; j++)
    DMat(LftChld - 1, j) = Frq(j) * tmpVec4(j);
  // cout << DMat << endl;
  for(int i = (numNds + numLvs - 2); i >= 0; i--)
  //  for(int i = 0; i < (numNds + numLvs - 1); i++)
    if(((i + 1) != rootLftChld) && ((i + 1) != rootRgtChld)){    
      Prnt = TreeMatrix[i][3]; 
      GrndPrnt = TreeMatrix[Prnt - 1][3]; GPBrLen = BranchLength(Prnt - 1);
      LftChld = TreeMatrix[Prnt - 1][1]; RgtChld = TreeMatrix[Prnt - 1][2];
      if(LftChld == (i + 1)) OthChld = RgtChld; else OthChld = LftChld;
      OthBrLen = BranchLength(OthChld - 1);
      
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
     }
 
    for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(OthChld - 1, j);
    tmpVec2 = RateToProbFun2(RateMat, OthBrLen, Frq, numSt) * tmpVec;
    
    for(int j = 0; j < numSt; j++) tmpVec3(j) = DMat(Prnt - 1, j);

    for(int j = 0; j < numSt; j++)
      for(int k = 0; k < numSt; k++)
	tmpVec4(j) += tmpVec3(k) * (RateToProbFun2(RateMat, GPBrLen, Frq, numSt))(k, j);

    for(int j = 0; j < numSt; j++)
      DMat(i, j) = tmpVec2(j)*tmpVec4(j);    
    }
  // cout << DMat << endl;

  /* Computing CndProbMat.... */
  for(int i = 0; i < (numNds + numLvs - 1); i++){
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
    }
    
    for(int j = 0; j < numSt; j++) 
      {tmpVec(j) = DMat(i, j); tmpVec2(j) = UMat(i, j);}
    
    count = 0;
    
    for(int j = 0; j < numSt; j++){
      for(int k = 0; k < numSt; k++){
	CndProbMat(i, count) = ((RateToProbFun2(RateMat, BranchLength(i), Frq, numSt))(k, j) * tmpVec(k) * tmpVec2(j))/logL;
	count++;
      }
    }
  }
  logL = log(logL);
}



LaGenMatDouble JFct(LaVectorDouble Lam, int numSt){
  LaGenMatDouble JMat(numSt, numSt);
  LaVectorDouble expLam(numSt);
  // cout << Lam << endl;
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++) JMat(i, j) = 0.0;
  for(int i = 0; i < numSt; i++) expLam(i) = 0.0;

  for(int i = 0; i < numSt; i++) expLam(i) = exp(Lam(i));
  // cout << expLam << endl;
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      if(abs(expLam(i)- expLam(j)) < 0.000000001){
     // if(expLam(i) == expLam(j)){
	JMat(i, j) = expLam(i);}
      else{
	JMat(i, j) = (expLam(i)-expLam(j))/(Lam(i)-Lam(j));
	//cout << i << " " << j << " " << Lam(i)<< " " << Lam(j)<< endl;
	}
    }
  
  return JMat;
}

LaGenMatDouble CndStatFct(LaVectorDouble Frq, LaSymmMatDouble Q, 
			  int numSt){
  LaGenMatDouble CndStat(numSt * numSt, numSt * numSt), V(numSt, numSt), 
    B(numSt, numSt), Diag(numSt, numSt), Diag2(numSt, numSt), BB(numSt, numSt),
    BBB(numSt, numSt), ProbQ(numSt, numSt), JMat(numSt, numSt), 
    Mat(numSt, numSt);
  LaVectorDouble SqrtFrq(numSt), Lam(numSt);
  long double ProbBE = 0.0, sum = 0.0, ColSum[numSt];
  int counter = 0;

  for(int i = 0; i < (numSt * numSt); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      CndStat(i, j) = 0.0;
  for(int i = 0; i < numSt; i++) SqrtFrq(i) = sqrt(Frq(i));

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++) {Diag(i, j) = 0.0;Diag2(i, j) = 0.0;}

  for(int i = 0; i < numSt; i++){
    Diag(i, i) = ((double)1)/SqrtFrq(i);
    ColSum[i] = 0.0;
  }

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      B(i, j) = SqrtFrq(i) * Q(i, j); ProbQ(i, j) = 0.0; JMat(i, j) = 0.0;
    }
  //  cout << Q << endl; exit(0);

  LaSymmMatDouble A(numSt,numSt);
  LaGenerateMatDouble(A);
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++)
      A(i,j) = (B * Diag)(i,j);
  //A(i,j) = B(i,j);
  //cout << A << endl;
  LaEigSolve(A, Lam, V);  
  //cout << Lam << endl << V << endl;
  for(int i = 0; i < numSt; i++)
    Diag2(i, i) = exp(Lam(i));
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      BB(i, j) = ((double)1)/SqrtFrq(i) * V(i, j);
      BBB(i, j) = SqrtFrq(i) * V(i, j);
    }

  ProbQ = (BB * Diag2) * Transposition(BBB, numSt);
  // cout << ProbQ << endl;

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++) JMat(i, j) = (JFct(Lam, numSt))(i, j);

  for(int BgnSt = 0; BgnSt < numSt; BgnSt++){
    for(int EndSt = 0; EndSt < numSt; EndSt++){
      ProbBE = ProbQ(BgnSt, EndSt);
      for (int i = 0; i < numSt; i++){
	sum = 0.0;
	for(int j = 0; j < numSt; j++) ColSum[j] = 0.0;
	for(int j = 0; j < numSt; j++)
	  for(int k = 0; k < numSt; k++){
	    Mat(j, k) = 0.0;
	    Mat(j, k) = V(i, j) * V(EndSt, j) * JMat(j, k);
	  }
	//cout << Mat << endl;
	for(int j = 0; j < numSt; j++) 
	  for(int k = 0; k < numSt; k++)
	    ColSum[j] += Mat(k, j);
	for(int j = 0; j < numSt; j++)
	  sum += V(BgnSt, j) * (V(i, j) * ColSum[j]);
	
	CndStat(i, numSt*(BgnSt)+EndSt) = ((SqrtFrq(EndSt)/SqrtFrq(BgnSt)) * sum)/ProbBE;
	//cout << numSt*(BgnSt)+EndSt << endl;
      }

      counter = numSt;

      for (int i = 0; i < numSt; i++){
	for(int j = 0; j < numSt; j++){
	  if(j != i){
	    sum = 0.0;
	    for(int k = 0; k < numSt; k++) ColSum[k] = 0.0;
	    for(int l = 0; l < numSt; l++)
	      for(int k = 0; k < numSt; k++){
		Mat(l, k) = 0.0;
		Mat(l, k) = V(j, l) * V(EndSt, l) * JMat(l, k);
	      }
	    for(int l = 0; l < numSt; l++)
	      for(int k = 0; k < numSt; k++)
		ColSum[l] += Mat(k, l);

	    for(int l = 0; l < numSt; l++)
	      sum +=  V(BgnSt, l) * (V(i, l) * ColSum[l]);

	    CndStat(counter, numSt * (BgnSt) + EndSt) = Q(i, j) * (SqrtFrq(i) * SqrtFrq(EndSt))/(SqrtFrq(BgnSt) * SqrtFrq(j)) * sum/ProbBE;
	    counter++;
	  }
	}
      }
    }
  }
  return CndStat;
}

LaGenMatDouble CndStatFct2(LaVectorDouble Frq, LaGenMatDouble Q, 
			  int numSt){
  LaGenMatDouble CndStat(numSt * numSt, numSt * numSt), V(numSt, numSt), 
    B(numSt, numSt), Diag(numSt, numSt), Diag2(numSt, numSt), BB(numSt, numSt),
    BBB(numSt, numSt), ProbQ(numSt, numSt), JMat(numSt, numSt), 
    Mat(numSt, numSt);
  LaVectorDouble SqrtFrq(numSt), Lam(numSt);
  long double ProbBE = 0.0, sum = 0.0, ColSum[numSt];
  int counter = 0;

  for(int i = 0; i < (numSt * numSt); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      CndStat(i, j) = 0.0;
  for(int i = 0; i < numSt; i++) SqrtFrq(i) = sqrt(Frq(i));

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++) {Diag(i, j) = 0.0;Diag2(i, j) = 0.0;}

  for(int i = 0; i < numSt; i++){
    Diag(i, i) = ((double)1)/SqrtFrq(i);
    ColSum[i] = 0.0;
  }

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      B(i, j) = SqrtFrq(i) * Q(i, j) * ((double)1)/SqrtFrq(j); ProbQ(i, j) = 0.0; JMat(i, j) = 0.0;
    }
  //cout << B << endl; 

  LaSymmMatDouble A(numSt,numSt);
  LaGenerateMatDouble(A);
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++)
      //A(i,j) = (B * Diag)(i,j);
      A(i,j) = B(i,j);
  //cout << A << endl;
  LaEigSolve(A, Lam, V);  
  //cout << Lam << endl << V << endl;
  for(int i = 0; i < numSt; i++)
    Diag2(i, i) = exp(Lam(i));
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      BB(i, j) = ((double)1)/SqrtFrq(i) * V(i, j);
      BBB(i, j) = SqrtFrq(i) * V(i, j);
    }

  ProbQ = (BB * Diag2) * Transposition(BBB, numSt);
  // cout << ProbQ << endl;

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++) JMat(i, j) = (JFct(Lam, numSt))(i, j);

  for(int BgnSt = 0; BgnSt < numSt; BgnSt++){
    for(int EndSt = 0; EndSt < numSt; EndSt++){
      ProbBE = ProbQ(BgnSt, EndSt);
      for (int i = 0; i < numSt; i++){
	sum = 0.0;
	for(int j = 0; j < numSt; j++) ColSum[j] = 0.0;
	for(int j = 0; j < numSt; j++)
	  for(int k = 0; k < numSt; k++){
	    Mat(j, k) = 0.0;
	    Mat(j, k) = V(i, j) * V(EndSt, j) * JMat(j, k);
	  }
	//cout << Mat << endl;
	for(int j = 0; j < numSt; j++) 
	  for(int k = 0; k < numSt; k++)
	    ColSum[j] += Mat(k, j);
	for(int j = 0; j < numSt; j++)
	  sum += V(BgnSt, j) * (V(i, j) * ColSum[j]);
	
	CndStat(i, numSt*(BgnSt)+EndSt) = ((SqrtFrq(EndSt)/SqrtFrq(BgnSt)) * sum)/ProbBE;
	//cout << numSt*(BgnSt)+EndSt << endl;
      }

      counter = numSt;

      for (int i = 0; i < numSt; i++){
	for(int j = 0; j < numSt; j++){
	  if(j != i){
	    sum = 0.0;
	    for(int k = 0; k < numSt; k++) ColSum[k] = 0.0;
	    for(int l = 0; l < numSt; l++)
	      for(int k = 0; k < numSt; k++){
		Mat(l, k) = 0.0;
		Mat(l, k) = V(j, l) * V(EndSt, l) * JMat(l, k);
	      }
	    for(int l = 0; l < numSt; l++)
	      for(int k = 0; k < numSt; k++)
		ColSum[l] += Mat(k, l);

	    for(int l = 0; l < numSt; l++)
	      sum +=  V(BgnSt, l) * (V(i, l) * ColSum[l]);

	    CndStat(counter, numSt * (BgnSt) + EndSt) = Q(i, j) * (SqrtFrq(i) * SqrtFrq(EndSt))/(SqrtFrq(BgnSt) * SqrtFrq(j)) * sum/ProbBE;
	    counter++;
	  }
	}
      }
    }
  }
  return CndStat;
}


void getEmpFreq(LaVectorDouble &FreqVec, int **SeqMat, int seqLn, int n,
		int numSt){
  int sum = 0;
  for(int i = 0; i < numSt; i++) FreqVec(i) = 0.0;
  for(int i = 0; i < n; i++)
    for(int j = 0; j < seqLn; j++){
      for(int k = 0; k < numSt; k++)
	if(SeqMat[i][j] == (k + 1)) FreqVec(k) += 1.0;
      sum++;
    }

  for(int i = 0; i < numSt; i++) FreqVec(i) /= ((double) sum);
  
}

void getEmpFreq(double **FreqVec, int **SeqMat, int seqLn, int n,
		int numSt){
  int sum = 0;
  for(int i = 0; i < numSt; i++) (*FreqVec)[i] = 0.0;
  for(int i = 0; i < n; i++)
    for(int j = 0; j < seqLn; j++){
      for(int k = 0; k < numSt; k++)
	if(SeqMat[i][j] == (k + 1)) (*FreqVec)[k] += 1.0;
      sum++;
    }

  for(int i = 0; i < numSt; i++) (*FreqVec)[i] /= ((double) sum);
  
}

/*
  int main(int argc, char **argv)
  {
  
  cout.precision(4);
  cout.setf(ios::scientific, ios::floatfield);
  int numSt = 4, numNds = 4, numLvs = 5;
  int Obs[numLvs], **PhlMat;
  LaVectorDouble BrLenV(numLvs + numNds - 1);
  double logL = 0.0;
  LaSymmMatDouble RateMat(numSt, numSt), Q(numSt, numSt);
  LaGenMatDouble CndProbMat(numNds + numLvs - 1, numSt * numSt);
  
  LaVectorDouble FreqVec(numSt);
  
  for(int i = 0; i < numSt; i++) FreqVec(i) = 0.25;
  
  PhlMat = new int*[numLvs + numNds];
  for(int i = 0; i < (numLvs + numNds); i++)
    PhlMat[i] = new int[4];
  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < 4; j++) 
      PhlMat[i][j] = 0;
  
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      RateMat(i, j) = ((double)1)/((double)3);
      Q(i, j) = ((double)1)/((double)6);
    }
  for(int i = 0; i < numSt; i++) {RateMat(i, i) = -1.0; Q(i, i) = -0.5;}
  
  PhlMat[0][0] = 1; PhlMat[1][0] = 1; PhlMat[2][0] = 1; PhlMat[3][0] = 1;
  PhlMat[4][0] = 1;
  PhlMat[0][3] = 6; PhlMat[1][3] = 6; PhlMat[2][3] = 7; PhlMat[3][3] = 7;
  PhlMat[4][3] = 8;
  
  PhlMat[5][1] = 1; PhlMat[5][2] = 2; PhlMat[5][3] = 9; PhlMat[6][1] = 3;
  PhlMat[6][2] = 4; PhlMat[6][3] = 8;
  PhlMat[7][1] = 7; PhlMat[7][2] = 5; PhlMat[7][3] = 9;
  PhlMat[8][1] = 6; PhlMat[8][2] = 8; PhlMat[8][3] = 0;
  
  Obs[0] = 1;  Obs[1] = 2; Obs[2] = 2; Obs[3] = 2; Obs[4] = 4;
  BrLenV(0) = 0.1;  BrLenV(1) = 0.1;  BrLenV(2) = 0.5;  BrLenV(3) = 0.5;
  BrLenV(4) = 0.1;  BrLenV(5) = 0.7;  BrLenV(6) = 0.1;  BrLenV(7) = 2.0;
  
  //   for(int i = 0; i < (numLvs + numNds); i++){
  //     for(int j = 0; j < 4; j++) cout << PhlMat[i][j] << " ";
  //     cout << endl;
  //   }
  //   LaGenMatDouble ProbMat(numSt, numSt);
  //   cout << RateToProbFun(RateMat, 0.5, FreqVec, numSt)<< endl;
  
  
  //exit(0);
  FlsnFct(PhlMat, RateMat, FreqVec, BrLenV, Obs, numSt, numLvs, numNds, CndProbMat, logL);
  
  // cout << CndProbMat<< endl << logL << endl;
  cout << CndStatFct(FreqVec, Q, numSt) << endl;
  return 0;
  }
*/

void computeEvalEvec(LaSymmMatDouble RateMat, LaGenMatDouble& InvFEgnVc,
		     LaGenMatDouble& TspFEgnVc, LaVectorDouble& EVal,
		     LaGenMatDouble& EigenRate,
		     LaVectorDouble FreqVec, int StateNum){
  LaGenMatDouble ProbMat(StateNum, StateNum);
  LaGenMatDouble Diag(StateNum, StateNum), EXP(StateNum, StateNum);
  LaGenMatDouble T(StateNum, StateNum),  B(StateNum, StateNum);
  LaVectorDouble Freq2(StateNum);
  
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++){
      TspFEgnVc(i, j) = 0.0; InvFEgnVc(i, j) = 0.0;
      ProbMat(i, j) = 0.0; Diag(i, j) = 0.0; EXP(i, j) = 0.0; 
      EigenRate(i, j) = 0.0; T(i, j) = 0.0; B(i, j) = 0.0;
    }
  for(int i = 0; i < StateNum; i++){
    EVal(i) = 0.0; Freq2(i) = 0.0;
  }

  for(int i = 0; i < StateNum; i++){
    FreqVec(i) = sqrt(FreqVec(i)); Diag(i, i) = 1/FreqVec(i);
    Freq2(i) = 1/FreqVec(i); }

  Mult(FreqVec, RateMat, B, StateNum);

  B = B * Diag;

  LaSymmMatDouble A(StateNum,StateNum);
  LaGenerateMatDouble(A);
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++)
      A(i,j) = B(i,j);

  LaEigSolve(A, EVal, EigenRate);  

  Mult(FreqVec, EigenRate, B, StateNum); 
  Transposition(B, TspFEgnVc, StateNum);
  Mult(Freq2, EigenRate, InvFEgnVc, StateNum); 

//   cout << fixed << InvFEgnVc << endl;
//   cout << endl << fixed << TspFEgnVc << endl;

}

void computeEvalEvec2(LaGenMatDouble RateMat, LaGenMatDouble& InvFEgnVc,
		     LaGenMatDouble& TspFEgnVc, LaVectorDouble& EVal,
		     LaGenMatDouble& EigenRate,
		     LaVectorDouble FreqVec, int StateNum){
  LaGenMatDouble ProbMat(StateNum, StateNum);
  LaGenMatDouble Diag(StateNum, StateNum), EXP(StateNum, StateNum);
  LaGenMatDouble T(StateNum, StateNum),  B(StateNum, StateNum);
  LaVectorDouble Freq2(StateNum);
  
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++){
      TspFEgnVc(i, j) = 0.0; InvFEgnVc(i, j) = 0.0;
      ProbMat(i, j) = 0.0; Diag(i, j) = 0.0; EXP(i, j) = 0.0; 
      EigenRate(i, j) = 0.0; T(i, j) = 0.0; B(i, j) = 0.0;
    }
  for(int i = 0; i < StateNum; i++){
    EVal(i) = 0.0; Freq2(i) = 0.0;
  }

  for(int i = 0; i < StateNum; i++){
    FreqVec(i) = sqrt(FreqVec(i)); Diag(i, i) = 1/FreqVec(i);
    Freq2(i) = 1/FreqVec(i); }

  Mult(FreqVec, RateMat, B, StateNum);

  B = B * Diag;

  //cout << B << endl;
  LaSymmMatDouble A(StateNum,StateNum);
  LaGenerateMatDouble(A);
  for(int i = 0; i < StateNum; i++)
    for(int j = 0; j < StateNum; j++)
      A(i,j) = B(i,j);

  LaEigSolve(A, EVal, EigenRate);  

  Mult(FreqVec, EigenRate, B, StateNum); 
  Transposition(B, TspFEgnVc, StateNum);
  Mult(Freq2, EigenRate, InvFEgnVc, StateNum); 

//   cout << fixed << InvFEgnVc << endl;
//   cout << endl << fixed << TspFEgnVc << endl;

}

void preFlsFct(int numSt, LaGenMatDouble InvFEgnVc, 
			 LaGenMatDouble TspFEgnVc, LaVectorDouble EgnVl,
			 double Br, LaGenMatDouble & tmpMat){
  LaVectorDouble tmpVec(numSt);
//   cout.precision(8);
//   cout << Br << endl;
  for(int i = 0; i < numSt; i++) 
    tmpVec(i) = 0.0;
  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++)
      tmpMat(i, j) = 0.0; 
  
  // LaGenMatDouble tmpMat(numSt, numSt);
  for(int j = 0; j < numSt; j++) tmpVec(j) = exp(Br*EgnVl(j));
  for(int j = 0; j < numSt; j++)
    for(int k = 0; k < numSt; k++)
      tmpMat(j, k) = InvFEgnVc(j, k) * tmpVec(k);
  
  tmpMat = tmpMat * TspFEgnVc;

  // cout << fixed << tmpMat << endl;

}

void FstFlsnFct(int** TreeMatrix, LaVectorDouble EgnVl, 
		LaGenMatDouble InvFEgnVc, 
		LaGenMatDouble TspFEgnVc, LaVectorDouble BranchLength, 
		LaVectorDouble Frq, int* Obs, int numSt, int numLvs, 
		int numNds, LaGenMatDouble & CndProbMat, long double & logL){
  int LftChld = 0, RgtChld = 0, rootLftChld = 0, rootRgtChld = 0, Prnt = 0,
    GrndPrnt = 0, OthChld = 0, count = 0;
  long double LftBrLen = 0.0, RgtBrLen = 0.0, GPBrLen = 0.0, 
    OthBrLen = 0.0;
  LaGenMatDouble UMat(numLvs + numNds, numSt), tmpMat(numSt, numSt), 
   tmpMat2(numSt, numSt), DMat(numLvs + numNds - 1, numSt);
  // CndProbMat(numNds + numLvs - 1, numSt * numSt); 
  LaVectorDouble tmpVec(numSt), tmpVec2(numSt), tmpVec3(numSt), tmpVec4(numSt);

  /* Initialize.... */
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < (numSt * numSt); j++)
      CndProbMat(i, j) = 0.0;

  for(int i = 0; i < (numLvs + numNds); i++)
    for(int j = 0; j < numSt; j++){
      UMat(i, j) = 0.0;
    }
  for(int i = 0; i < (numLvs + numNds - 1); i++)
    for(int j = 0; j < numSt; j++){
      DMat(i, j) = 0.0;
    }

  for(int i = 0; i < numSt; i++)
    for(int j = 0; j < numSt; j++){
      tmpMat(i, j) = 0.0; tmpMat2(i, j) = 0.0;
    }

  for(int i = 0; i < numSt; i++) {
    tmpVec(i) = 0.0; tmpVec2(i) = 0.0; tmpVec3(i) = 0.0; tmpVec4(i) = 0.0; 
  }
  logL = 0.0;

  /* U matrix.... */
  for(int i = 0; i < numLvs; i++)
    UMat(i, Obs[i] - 1) = 1;
  for(int i = 0; i < numNds; i++){
    LftChld = TreeMatrix[numLvs + i][1];
    RgtChld = TreeMatrix[numLvs + i][2];
    LftBrLen = BranchLength(LftChld - 1); RgtBrLen = BranchLength(RgtChld - 1);

    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
    }
  
    preFlsFct(numSt, InvFEgnVc, TspFEgnVc, EgnVl, LftBrLen, tmpMat);

    for(int j = 0; j < numSt; j++)
      tmpVec(j) = UMat(LftChld - 1, j);

    tmpVec3 = tmpMat * tmpVec;

    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0;
    }

    preFlsFct(numSt, InvFEgnVc, TspFEgnVc, EgnVl, RgtBrLen, tmpMat2);

    for(int j = 0; j < numSt; j++)
      tmpVec(j) = UMat(RgtChld - 1, j);

    tmpVec4 = tmpMat2 * tmpVec;

    for(int j = 0; j < numSt; j++)
      UMat(numLvs + i, j) = tmpVec3(j) * tmpVec4(j);

  }

  //cout << UMat << endl;
 
  /* log L */
  for(int j = 0; j < numSt; j++)
    logL += UMat(numNds + numLvs - 1, j) * Frq(j);

  // cout << logL << endl;
  /* D matrix..... */

  LftChld = TreeMatrix[numLvs + numNds - 1][1]; 
  LftBrLen = BranchLength(LftChld - 1);
  RgtChld = TreeMatrix[numLvs + numNds - 1][2]; 
  RgtBrLen = BranchLength(RgtChld - 1);

  rootLftChld = LftChld; rootRgtChld = RgtChld;

  // cout << LftChld << " " << RgtChld << endl;
  for(int j = 0; j < numSt; j++){
    tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
  }

  for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(LftChld - 1, j);
  tmpVec2 = tmpMat * tmpVec;
  
  for(int j = 0; j < numSt; j++)
    DMat(RgtChld - 1, j) = Frq(j) * tmpVec2(j);

  for(int j = 0; j < numSt; j++) tmpVec3(j) = UMat(RgtChld - 1, j);
  tmpVec4 = tmpMat2 * tmpVec3;

  for(int j = 0; j < numSt; j++)
    DMat(LftChld - 1, j) = Frq(j) * tmpVec4(j);
  //  cout << DMat << endl; exit(0);
  for(int i = (numNds + numLvs - 2); i >= 0; i--)
  //  for(int i = 0; i < (numNds + numLvs - 1); i++)
    if(((i + 1) != rootLftChld) && ((i + 1) != rootRgtChld)){    
      Prnt = TreeMatrix[i][3]; 
      GrndPrnt = TreeMatrix[Prnt - 1][3]; GPBrLen = BranchLength(Prnt - 1);
      LftChld = TreeMatrix[Prnt - 1][1]; RgtChld = TreeMatrix[Prnt - 1][2];
      if(LftChld == (i + 1)) OthChld = RgtChld; else OthChld = LftChld;
      OthBrLen = BranchLength(OthChld - 1);
      
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; tmpVec3(j) = 0.0; tmpVec4(j) = 0.0;
     }
 

    for(int j = 0; j < numSt; j++) tmpVec(j) = UMat(OthChld - 1, j);

    preFlsFct(numSt, InvFEgnVc, TspFEgnVc, EgnVl, OthBrLen, tmpMat);

    tmpVec2 = tmpMat * tmpVec;
    
    for(int j = 0; j < numSt; j++) tmpVec3(j) = DMat(Prnt - 1, j);

   preFlsFct(numSt, InvFEgnVc, TspFEgnVc, EgnVl, GPBrLen, tmpMat2);

    for(int j = 0; j < numSt; j++)
      for(int k = 0; k < numSt; k++)
	tmpVec4(j) += tmpVec3(k) * tmpMat2(k, j);

    for(int j = 0; j < numSt; j++)
      DMat(i, j) = tmpVec2(j) * tmpVec4(j);    
    }
  // cout << DMat << endl; exit(0);

  /* Computing CndProbMat.... */
  for(int i = 0; i < (numNds + numLvs - 1); i++){
    for(int j = 0; j < numSt; j++){
      tmpVec(j) = 0.0; tmpVec2(j) = 0.0; 
    }
    
    for(int j = 0; j < numSt; j++) 
      {tmpVec(j) = DMat(i, j); tmpVec2(j) = UMat(i, j);}
    
    count = 0;
   
    for(int j = 0; j < numSt; j++){
      for(int k = 0; k < numSt; k++){
	preFlsFct(numSt, InvFEgnVc, TspFEgnVc, EgnVl, BranchLength(i), tmpMat2);
	CndProbMat(i, count) = ((tmpMat2(k, j) * tmpVec(k)) * tmpVec2(j))/logL;
	count++;
      }
    }
  }
  logL = log(logL);
}
