/*********************************************************************
 Author:  Ruriko Yoshida
 Date:    March 8th, 2004
 Update:  February 24th, 2005
 
Comments: This file contains reading function from MFA file and 
          converting to NJ file.
***********************************************************************/
#include "Timer.h"
#include "Reading.h"
#include "Subset.h"
#include "Taxa.h"
#include "fastDNAmlheader.h"
#include "fastDNAmlModified2.h"


/*************** getList ***********************************************/

void Subsets3(int N, int M, int I, int K, int *array, int ** out, int & counter)
{
    //If we have already excluded too many items, then return
    int j = 0;
    if((M-K) == (N-I+1))
         return;
    
    
    //if we have finished comuting a subset, then print it out
    
    if((M-K) == 0)
    {
        out[counter] = new int[M];
        for(int i = 0; i < N; i++)
	  if(array[i] == 1){
                 out[counter][j] = i;
		 j++;
	  }
	counter++;
         return;
     }
 
 
        
    int newArray[N];
    for(int i = 0; i < N; i++) newArray[i] = 0;
 
    //copy the K precomputed values from array[]
 
    for(int i = 0; i < I; i++)
        newArray[i] = array[i];
 
 
     //compute all subsets where we choose element I
     newArray[I] = 1;
     Subsets3(N, M, I+1, K+1, newArray, out, counter);
 
     //compute all subsets where we DON'T choose element I
     newArray[I] = 0;
     Subsets3(N, M, I+1, K, newArray, out, counter);
 
    return;
    
 
}
 
 
int getList3(int N, int M, int** out){

  int array[N], counter = 0;

  for(int i = 0; i < N; i++) array[i] = 0;
  //Print out all subsets of size M
 
  Subsets3(N, M, 0, 0, array, out, counter);

  return 0;
}


/************************************************************************/

void checkMfaFile(char* filename){
  string name;
  char test;

  ifstream in(filename);
  if(!in){
    cerr << "Input file is missing!" << endl;
    exit(0);
  }
  while(in >> test){
    if(test == '>') in >> name;
    else  if(test == 'A') ;
    else  if(test == 'B') ;
    else  if(test == 'C') ;
    else  if(test == 'D') ;
    else  if(test == 'G') ;
    else  if(test == 'H') ;
    else  if(test == 'K') ;
    else  if(test == 'M') ;
    else  if(test == 'N') ;
    else  if(test == 'O') ;
    else  if(test == 'R') ;
    else  if(test == 'S') ;
    else  if(test == 'T') ;
    else  if(test == 'U') ;
    else  if(test == 'V') ;
    else  if(test == 'W') ;
    else  if(test == 'X') ;
    else  if(test == 'Y') ;
    else  if(test == '?') ;
    else  if(test == '-') ;
    else {
      cerr << "Input file contains charactor " << test << "." << endl;
      cerr << " Please check the input file!" << endl;
      exit(0);
    }
  }
}


void WriteFafsaToPhylipNJ(int n, int m, Alinements* InputData, int* index){
  ofstream out;
  unsigned int length = InputData[index[0]].DNA.length();
  unsigned int max = 0;
  //  for(int i = 0; i < m; i++)
  //if(max < strlen(InputData[index[i]].name)) max = strlen(InputData[index[i]].name);
  //  cout << max << endl;
  max += 10;

  out.open("out.test");
  if(m < 4) m = 4;
  //   out << 4 << " ";
  out << m << " ";
  out << length << endl;

  for(int i = 0; i < m; i++)
    {
      //  for(unsigned int j = 0; j < strlen(InputData[index[i]].name); j++)
      // out << InputData[index[i]].name[j];
      out << "S" << i;

      for(unsigned int j = 0; j < max; j++)
	out << " ";
      out << InputData[index[i]].DNA << endl;
      if(InputData[index[i]].DNA.length() < length)
	for(unsigned int j = 0; j < (length - InputData[index[i]].DNA.length()); j++)
	  out << "-" << endl;
    }
}

void WriteFafsaToPhylip(int n, int m, Alinements* InputData, int* index){
  ofstream out;
  unsigned int length = InputData[index[0]].DNA.length();
  unsigned int max = 0;
  for(int i = 0; i < m; i++)
    if(max < strlen(InputData[index[i]].name)) max = strlen(InputData[index[i]].name);
  //  cout << max << endl;
  max += 10;

  out.open("out.test");
  if(m < 4) m = 4;
  //   out << 4 << " ";
  out << m << " ";
  out << length << endl;

  for(int i = 0; i < m; i++)
    {
      for(unsigned int j = 0; j < strlen(InputData[index[i]].name); j++)
      out << InputData[index[i]].name[j];

      for(unsigned int j = 0; j < (max - strlen(InputData[index[i]].name)); j++)
	out << " ";
      out << InputData[index[i]].DNA << endl;
      if(InputData[index[i]].DNA.length() < length)
	for(unsigned int j = 0; j < (length - InputData[index[i]].DNA.length()); j++)
	  out << "-" << endl;
    }
}


void StringToChar(int n, int m, Alinements* InputData, int* index, char** dna, int & length){
  ofstream out;
  length = (int) InputData[index[0]].DNA.length();

  for(int i = 0; i < m; i++)
    {
      //      dna[i] = new char[length + 1];
      strcpy(dna[i], InputData[index[i]].DNA.c_str()); 
      //cout << dna[i] << endl;
    } //exit(0);
}
 

void ReadFMAFile(int & n, int & m, char* filename){
   int counter = 0;
    string tmpString;
    double w[3], Wts[m];
    for(int i = 0; i < 3; i++) w[i] = 0;
    for(int i = 0; i < m; i++) Wts[i] = 0.000001;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    ifstream in(filename);
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> jnk;
    if(jnk == '>') {
      in >> InputData[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	InputData[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    InputData[counter - 1].number = counter;
    in.putback(jnk);
    }
    
    
    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }

   if(m == 2){
     int flag1 = 0, flag2 = 0, INDEX[4];
     double hold = 0, s = 0, zero = 0;
     char tmp[127];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	  ofstream OUT("dummy.out");
	  
// 	  if(!OUT){
// 	    cerr << "Error on writing dummy.out." << endl;
// 	    exit(1);
// 	    }
	  
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
	  //  system("./fasta2phylip dummy.out out.mfa");
	  system("./fastDNAml < out.test > out.tax");
	  
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	    
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) 
	    IN >> tmpString;
	  
	  IN >> tmp;
	  w[0] = atof(tmp);
	  
	  // cout << tmp << endl;
	    
	  if(flag1 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[0]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[0]; flag2 = 1;}

	  if(w[0] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[0];
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;

	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[1]; flag2 = 1;}
	  
	  if(w[1] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}
	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 

	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 
	    
	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    out.precision(10);
	    if(hold != 0) 
	      for(int i = 0; i < m; i++) Wts[i] = hold /2;

	    out << i + 1 << " " << j + 1 << " " << w[0] << " " << Wts[0] << " " << Wts[1] << endl;
	    s += w[0];
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) Wts[i] = 0.00001;
	    flag1 = 0; flag2 = 0;
	}
      }
   }

    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4];
      double hold = 0, s = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) flags[i] = 0;
      for(int i = 0; i < m; i++) Wts[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	    ofstream OUT("dummy.out");
	    
// 	    if(!OUT){
// 	      cerr << "Error on writing dummy.out." << endl;
// 	      exit(1);
// 	    }
	    
// 	    zero = 0;

// 	    for(int i = 0; i < m; i++){
// 	      OUT << ">" << InputData[index[i]].name << endl;
// 	      OUT << InputData[index[i]].DNA << endl;
// 	    }

// 	    OUT << ">" << InputData[index[2]].name << endl;
// 	    OUT << InputData[index[2]].DNA << endl;

	    // system("./fasta2phylip dummy.out out.mfa");
	    system("./fastDNAml < out.test > out.tax");
	    
	    ifstream IN("out.tax");
	    if(!IN){
	      cerr << "Error on reading out.tax." << endl;
	      exit(1);
	    }
	    
	    while(tmpString != "Limits") {
	      IN >> tmpString; 
	    }
	    for(int p = 0; p < 8; p++) 
	      IN >> tmpString;


	    IN >> tmp;
	    w[0] = atof(tmp);

	    if(w[0] != 0){
	      int sums = 0;
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name)
		  if(flags[i] == 0){
		    Wts[i] = w[0];
		    flags[i] = 1;
		  }
		sums += flags[i];
	      }
	      if(sums == 0) hold = w[0];
	    }

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;
	    if((tmpString == "**")|| (tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0;

	    IN >> tmp;
	    w[1] = atof(tmp);

	    w[0] += w[1];

	    if(w[1] != 0){
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name){
		  if(flags[i] == 0){
		    Wts[i] = w[1];
		    flags[i] = 1;
		  }
		  flag = 1;
		}
		if(flag == 0) hold = w[1]; 
	      }
	    }

	    flag = 0;

	    for(int j2 = 0; j2 < 2 * m - 3; j2++){
	      for(int p = 0; p < 4; p++) 
		IN >> tmpString;
	      if((tmpString == "**")|| (tmpString == "*"))
		zero = 1;
	      if(zero == 1) for(int p = 0; p < 2; p++) {
		zero = 0;
		IN >> tmpString;
	    }
	      
	      else 
		IN >> tmpString;
	      zero = 0;
	  
	      IN >> tmp;
	      w[1] = atof(tmp);

	      w[0] += w[1];

	      if(w[1] != 0){
		for(int i = 0; i < m; i++){
		  if(tmpString == InputData[index[i]].name){
		    if(flags[i] == 0){
		      Wts[i] = w[1];
		      flags[i] = 1;
		    }
		    flag = 1;
		  }
		  if(flag == 0) hold = w[1];
		}		 
	      }
	      flag = 0;
	    }

	      Wts[2] = hold;

	    out << index[0] + 1 << " " << index[1] + 1 << " " << index[2] + 1 << " " << w[0] << " ";
	    for(int i = 0; i < m; i++){
	      if(Wts[i] != 0) out << Wts[i] << " ";
	      else out << "0.000001 ";
	    }
	    out << endl;
	    s += w[0];
	    flag = 0;
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) flags[i] = 0;
	    for(int i = 0; i < m; i++) Wts[i] = 0;
	    hold = 0;
	    
	  }
	}
      }

    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) Wts[i] = 0;
      WW = numOfWeight;

      int index[WW][m];

      ifstream INDEX;
      getList(n, m);

      INDEX.open("ListOfSubsets");
      if(!INDEX){
	cerr<<"Error for reading ListOfSubsets." << endl;
	exit(0);
      }
      for(int i = 0; i < numOfWeight; i++)
	for(int j = 0; j < m; j++) INDEX >> index[i][j];
      for(int i = 0; i < numOfWeight; i++){
	  WriteFafsaToPhylip(n, m, InputData, index[i]);
// 	ofstream OUT("dummy.out");
	
// 	if(!OUT){
// 	  cerr << "Error on writing dummy.out." << endl;
// 	  exit(1);
// 	}
// 	for(int j = 0; j < m; j++){	    
// 	  OUT << ">" << InputData[index[i][j]].name << endl;
// 	  OUT << InputData[index[i][j]].DNA << endl;
// 	}
// 	system("./fasta2phylip dummy.out out.mfa");
	system("./fastDNAml < out.test > out.tax");
	
	ifstream IN("out.tax");
	if(!IN){
	  cerr << "Error on reading out.tax." << endl;
	  exit(1);
	}
	
	while(tmpString != "Limits") {
	  IN >> tmpString; 
	}
	for(int p = 0; p < 8; p++) IN >> tmpString;
	
	IN >> tmp;
	w[0] = atof(tmp);
	
	//cout << tmp << endl;

	if(w[0] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[0];
	  }
	}
      
	for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	if((tmpString == "**")|| (tmpString == "*"))zero = 1;
	if(zero == 1) for(int p = 0; p < 2; p++) {
	  zero = 0;
	  IN >> tmpString;
	}
	
	else 
	  IN >> tmpString;
	zero = 0;

	IN >> tmp;
	w[1] = atof(tmp);
	//cout << tmp << endl;
	w[0] += w[1];	
	
	if(w[1] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[1];
	  }
	}
	
	for(int j2 = 0; j2 < 2 * m - 5; j2++){
	  
	  for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	  if((tmpString == "**")|| (tmpString == "*"))
	    zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0;
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  //cout << tmp << endl;

	  w[0] += w[1];	
	  
	  if(w[1] != 0){
	    for(int j = 0; j < m; j++){
	      if(tmpString == InputData[index[i][j]].name)
		Wts[j] = w[1];
	    }	//cout << tmp << endl;
	  }
	}
	
	for(int j = 0; j < m; j++)   
	  out << index[i][j] + 1 << " ";
	out << w[0] << " ";
	
	for(int i = 0; i < m; i++){
	  if(Wts[i] != 0) out << Wts[i] << " ";
	  else out << "0.000001 ";
	}
	out << endl;
	s += w[0];
	w[0] = 0;
	w[1] = 0;
	w[2] = 0;
	for(int i = 0; i < m; i++) Wts[i] = 0;
      }
      system ("rm ListOfSubsets");
      //  cout << endl << s << endl;
    }
    
    // system ("rm out.mfa");
    system ("rm checkpoint.*");
    system ("rm treefile.*");
   
}


/****************************************************************/
void ReadFMAFileNJ(int & n, int & m, char* filename){
   int counter = 0;
    string tmpString;
    double w[3];
    for(int i = 0; i < 3; i++) w[i] = 0;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    ifstream in(filename);
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> jnk;
    if(jnk == '>') {
      in >> InputData[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	InputData[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    InputData[counter - 1].number = counter;
    in.putback(jnk);
    }
    
    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }

   if(m == 2){
     int INDEX[4];
     double s = 0, zero = 0;
     char tmp[127];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  WriteFafsaToPhylipNJ(n, m, InputData, INDEX);
// 	  ofstream OUT("dummy.out");
	  
// 	  if(!OUT){
// 	    cerr << "Error on writing dummy.out." << endl;
// 	    exit(1);
// 	    }
	  
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
	  //  system("./fasta2phylip dummy.out out.mfa");
	  system("./fastDNAml < out.test > out.tax");
	  
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	    
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) 
	    IN >> tmpString;
	  
	  IN >> tmp;
	  w[0] = atof(tmp);
	  
	  // cout << tmp << endl;
	    
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;

	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 

	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];


	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	 
	  out.precision(10);
	  
	  out << i + 1 << " " << j + 1 << " " << w[0] << endl;
	    s += w[0];
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

      system ("rm checkpoint.*");
      system ("rm treefile.*");

	}
      }
   }

    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4];
      double s = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) flags[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  WriteFafsaToPhylipNJ(n, m, InputData, INDEX);
	  
	  system("./fastDNAml < out.test > out.tax");
	  
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	  
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) 
	    IN >> tmpString;
	  
	  
	  IN >> tmp;
	  w[0] = atof(tmp);
	  
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  if((tmpString == "**")|| (tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0;
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  w[0] += w[1];
	  
	  for(int j2 = 0; j2 < 2 * m - 3; j2++){
	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;
	    if((tmpString == "**")|| (tmpString == "*"))
	      zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }
	      
	      else 
		IN >> tmpString;
	      zero = 0;
	  
	      IN >> tmp;
	      w[1] = atof(tmp);

	      w[0] += w[1];

	  }


	    out << index[0] + 1 << " " << index[1] + 1 << " " << index[2] + 1 << " " << w[0] << " ";

	    out << endl;
	    s += w[0];
	    flag = 0;
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;
    system ("rm checkpoint.*");
    system ("rm treefile.*");

	  }
	}
      }      
    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127];

      WW = numOfWeight;

      int index[WW][m];

      ifstream INDEX;
      getList(n, m);

      INDEX.open("ListOfSubsets");
      if(!INDEX){
	cerr<<"Error for reading ListOfSubsets." << endl;
	exit(0);
      }
      for(int i = 0; i < numOfWeight; i++)
	for(int j = 0; j < m; j++) INDEX >> index[i][j];
      for(int i = 0; i < numOfWeight; i++){
	WriteFafsaToPhylipNJ(n, m, InputData, index[i]);

	system("./fastDNAml < out.test > out.tax");
	
	ifstream IN("out.tax");
	if(!IN){
	  cerr << "Error on reading out.tax." << endl;
	  exit(1);
	}
	
	while(tmpString != "Limits") {
	  IN >> tmpString; 
	}
	for(int p = 0; p < 8; p++) IN >> tmpString;
	
	IN >> tmp;
	w[0] = atof(tmp);
	
	//cout << tmp << endl;

      
	for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	if((tmpString == "**")|| (tmpString == "*"))zero = 1;
	if(zero == 1) for(int p = 0; p < 2; p++) {
	  zero = 0;
	  IN >> tmpString;
	}
	
	else 
	  IN >> tmpString;
	zero = 0;

	IN >> tmp;
	w[1] = atof(tmp);
	//cout << tmp << endl;
	w[0] += w[1];	
	
	
	for(int j2 = 0; j2 < 2 * m - 5; j2++){
	  
	  for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	  if((tmpString == "**")|| (tmpString == "*"))
	    zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0;
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  //cout << tmp << endl;

	  w[0] += w[1];	
	  
	}
	
	for(int j = 0; j < m; j++)   
	  out << index[i][j] + 1 << " ";
	out << w[0] << " ";
	
	out << endl;
	s += w[0];
	w[0] = 0;
	w[1] = 0;
	w[2] = 0;

	system ("rm checkpoint.*");
	system ("rm treefile.*");

      }
      system ("rm ListOfSubsets");
      //  cout << endl << s << endl;
    }
    
    // system ("rm out.mfa");
//     system ("rm checkpoint.*");
//     system ("rm treefile.*");
   
}

/***************************************************************/

void ReadFMAFileNJ2(int & n, int & m, char* filename){
   int counter = 0;
    string tmpString;
    double w[3];
    for(int i = 0; i < 3; i++) w[i] = 0;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    ifstream in(filename);
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> jnk; 
    if(jnk == '>') {
      in >> InputData[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	InputData[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    InputData[counter - 1].number = counter;
    in.putback(jnk);
    }
    
    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }

   if(m == 2){
     int INDEX[4], length = 0;
     double s = 0, zero = 0;
     char tmp[127], *dna[4];
     for(int i = 0; i < 4; i++) 
       dna[i] = new char[InputData[0].DNA.length() + 1];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  StringToChar(n, 4, InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << i + 1 << " " << j + 1 << " " << getWts(4, length, dna) << endl;

	}
      }
     for(int i = 0; i < 4; i++) 
       delete [] dna[i];
   }

  
    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4], length = 0;
      double s = 0;
      char tmp[127], *dna[4];
      for(int i = 0; i < 4; i++) 
	dna[i] = new char[InputData[0].DNA.length() + 1];

      for(int i = 0; i < m; i++) flags[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  StringToChar(n, 4, InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << index[0] + 1 << " " << index[1] + 1 << " " << index[2] + 1 << " " << getWts(4, length, dna) << endl;

	  }
	}
      }
      for(int i = 0; i < 4; i++) 
	delete [] dna[i];
    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127], *dna[m];
      for(int i = 0; i < m; i++) 
	dna[i] = new char[InputData[0].DNA.length() + 1];

      WW = numOfWeight;

      int index[WW][m], length = 0;

      ifstream INDEX;
      getList(n, m);

      INDEX.open("ListOfSubsets");
      if(!INDEX){
	cerr<<"Error for reading ListOfSubsets." << endl;
	exit(0);
      }
      for(int i = 0; i < numOfWeight; i++)
	for(int j = 0; j < m; j++) INDEX >> index[i][j];
      for(int i = 0; i < numOfWeight; i++){
	StringToChar(n, m, InputData, index[i], dna, length);
	
	for(int j = 0; j < m; j++)   
	  out << index[i][j] + 1 << " ";
	out << getWts(m, length, dna)<< " ";
	
	out << endl;
	
      }
      for(int i = 0; i < m; i++) 
	delete [] dna[i];
      system("rm ListOfSubsets");
    }
 
}


/* subset function */

void Subsets4(int N, int M, int I, int K, int *array, int & counter, 
	      char* freq, char* TsTv, double tfreq, double cfreq, 
	      double afreq, double gfreq, double trans,  
	      Alinements* InputData, char** dna, int & length, int & num,
	      int numOfWeight, ofstream & OUT)
  
{
  //If we have already excluded too many items, then return
  int j = 0;
  if((M-K) == (N-I+1))
    return;
  
  
  //if we have finished comuting a subset, then print it out
  
    if((M-K) == 0)
      {
        int * out;
	out = new int[M];
        for(int i = 0; i < N; i++)
	  if(array[i] == 1){
	    out[j] = i;
		 j++;
	  }
	/* New function for computing m-wts. */
	StringToChar(N, M, InputData, out, dna, length);
	if(M != N){
	  OUT.precision(10);
	  OUT << fixed << getWts(M, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< " ";
	  num++;
	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	}
	else {
	  exit(0);
	}
	delete [] out;
	/* end */
	counter++;
         return;
     }
 
 
        
    int newArray[N];
    for(int i = 0; i < N; i++) newArray[i] = 0;
 
    //copy the K precomputed values from array[]
 
    for(int i = 0; i < I; i++)
        newArray[i] = array[i];
 
 
     //compute all subsets where we choose element I
     newArray[I] = 1;
     Subsets4(N, M, I+1, K+1, newArray, counter, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, InputData, dna, length, num, numOfWeight, OUT);
 
     //compute all subsets where we DON'T choose element I
     newArray[I] = 0;
     Subsets4(N, M, I+1, K, newArray, counter, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, InputData, dna, length, num, numOfWeight, OUT);
 
    return;
    
 
}
 
 
int getList4(int N, int M, char* freq, char* TsTv, double tfreq, double cfreq, 
	     double afreq, double gfreq, double trans, Alinements* InputData, char** dna, int length,
	     int & num, int numOfWeight, ofstream & OUT){

  int array[N], counter = 0;

  for(int i = 0; i < N; i++) array[i] = 0;
  //Print out all subsets of size M
 
  Subsets4(N, M, 0, 0, array, counter, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, InputData, dna, length, num, numOfWeight, OUT);

  return 0;
}

/* end */


void ReadFMAFileNJ3(int & n, int & m, char* filename, char* freq, char* TsTv, 
                    double tfreq, double cfreq, double afreq, double gfreq, 
		    double trans){
   int counter = 0, num = 0;
   string tmpString;
   double w[3];
    for(int i = 0; i < 3; i++) w[i] = 0;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    // cout << "Reading the input file...";
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    if(n < m+1){
      cerr << "Error: Please check that m < n where n is the number of taxa." << endl << endl;
      exit(0);
    }


    ifstream in(filename);
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> jnk; 
    if(jnk == '>') {
      in >> InputData[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	InputData[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    InputData[counter - 1].number = counter;
    in.putback(jnk);
    }
    
    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }


   if(m == 2){
     int INDEX[4], length = 0;
     double s = 0, zero = 0;
     char tmp[127], *dna[4];
     for(int i = 0; i < 4; i++) 
       dna[i] = new char[InputData[0].DNA.length() + 1];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  StringToChar(n, 4, InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	}
      } out << endl;
     for(int i = 0; i < 4; i++) 
       delete [] dna[i];
   }

  
    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4], length = 0;
      double s = 0;
      char tmp[127], *dna[4];
      for(int i = 0; i < 4; i++) 
	dna[i] = new char[InputData[0].DNA.length() + 1];

      for(int i = 0; i < m; i++) flags[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  StringToChar(n, 4, InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	  }
	}
      }
      out << endl;
      for(int i = 0; i < 4; i++) 
	delete [] dna[i];
    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127], *dna[m];
      for(int i = 0; i < m; i++) 
	dna[i] = new char[InputData[0].DNA.length() + 1];
      
      WW = numOfWeight;

      int length = 0;// *index[WW];
//       for(int i = 0; i < numOfWeight; i++) 
// 	index[i] = new int[m];

//      getList3(n, m, index);
      getList4(n, m, freq, TsTv, tfreq, cfreq, 
	 afreq, gfreq, trans, InputData, dna, length,
	 num, numOfWeight, out);

//       for(int i = 0; i < numOfWeight; i++){
// 	StringToChar(n, m, InputData, index[i], dna, length);
// 	if(m != n){
// 	  out << fixed << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< " ";
// 	  num++;
// 	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
// 	}
// 	else {
// 	  //	  ofstream total("totalWeight");
// 	  // total << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< endl;
// 	  exit(0);
// 	}

// 	delete [] index[i];
//       }
      out << endl;
      for(int i = 0; i < m; i++) 
	delete [] dna[i];
    }
    
}


/********** EMGNJ *************************/

void ReadFMAFileNJ4(int & n, int & m, char* filename, char* freq, char* TsTv, 
                    double &tfreq, double &cfreq, double &afreq, double &gfreq,
		    double trans, Alinements ** InputData){
   int counter = 0, num = 0;
   string tmpString;
   double w[3];
    for(int i = 0; i < 3; i++) w[i] = 0;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    // cout << "Reading the input file...";
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    if(n < m+1){
      cerr << "Error: Please check that m < n where n is the number of taxa." << endl << endl;
      exit(0);
    }


    ifstream in(filename);
    
    *InputData = new Alinements[n];
    for(int i = 0; i < n; i++){
      in >> jnk; 
    if(jnk == '>') {
      in >> (*InputData)[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	(*InputData)[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    (*InputData)[counter - 1].number = counter;
    in.putback(jnk);
    }
    


    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << (*InputData)[i].name << endl;
    }

   if(m == 2){
     int INDEX[4], length = 0;
     double s = 0, zero = 0;
     char tmp[127], *dna[4];
     for(int i = 0; i < 4; i++) 
       dna[i] = new char[(*InputData)[0].DNA.length() + 1];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  StringToChar(n, 4, *InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	}
      } out << endl;
     for(int i = 0; i < 4; i++) 
       delete [] dna[i];
   }

  
    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4], length = 0;
      double s = 0;
      char tmp[127], *dna[4];
      for(int i = 0; i < 4; i++) 
	dna[i] = new char[(*InputData)[0].DNA.length() + 1];

      for(int i = 0; i < m; i++) flags[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  StringToChar(n, 4, *InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	  }
	}
      }
      out << endl;
      for(int i = 0; i < 4; i++) 
	delete [] dna[i];
    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127], *dna[m];
      for(int i = 0; i < m; i++) 
	dna[i] = new char[(*InputData)[0].DNA.length() + 1];
      
      WW = numOfWeight;

      int length = 0;// *index[WW];
//       for(int i = 0; i < numOfWeight; i++) 
// 	index[i] = new int[m];

//      getList3(n, m, index);
      getList4(n, m, freq, TsTv, tfreq, cfreq, 
	 afreq, gfreq, trans, *InputData, dna, length,
	 num, numOfWeight, out);

//       for(int i = 0; i < numOfWeight; i++){
// 	StringToChar(n, m, InputData, index[i], dna, length);
// 	if(m != n){
// 	  out << fixed << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< " ";
// 	  num++;
// 	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
// 	}
// 	else {
// 	  //	  ofstream total("totalWeight");
// 	  // total << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< endl;
// 	  exit(0);
// 	}

// 	delete [] index[i];
//       }
      out << endl;
      for(int i = 0; i < m; i++) 
	delete [] dna[i];
    }
    
}

/***************************************************************/

void ReadTaxaNJ(Taxa * Data, int n, int m, int numOfWeights, ifstream & in, char ** names){
  char tmp[127];

  for(int i = 0; i < n; i++){
    in >> tmp;
    names[i] = new char[strlen(tmp) + 1];
    strcpy(names[i], tmp);
  }
  
  if(n <  m){
    cerr << "The number of leaves must be more than m. " << endl;
    exit(0);
  }
 
  for(int f = 0; f < numOfWeights; f++){
    for(int i = 0; i < m; i++) in >> Data[f].array[i];
    in >> Data[f].wt;// cout << Data[f].wt << endl;
    delete [] Data[f].edgeWts;
    //    for(int i = 0; i < m; i++) in >> Data[f].edgeWts[i];
  }

} 

void ReadTaxaNJ2(Taxa * Data, int n, int m, long long int numOfWeights, ifstream & in, char ** names){
  char tmp[127];

  for(int i = 0; i < n; i++){
    in >> tmp;
    names[i] = new char[strlen(tmp) + 1];
    strcpy(names[i], tmp);
  }
  
  if(n < m){
    cerr << "The number of leaves must be more than m. " << endl;
    exit(0);
  }
 
  for(long long int f = 0; f < numOfWeights; f++){
    delete [] Data[f].array;
    in >> Data[f].wt;// cout << Data[f].wt << endl;
    delete [] Data[f].edgeWts;
    //    for(int i = 0; i < m; i++) in >> Data[f].edgeWts[i];
  }

} 

/****************************************************************************/
void ReadFMAFile2(int & n, int & m, char* filename){
   int counter = 0;
    string tmpString;
    double w[3], Wts[m];
    for(int i = 0; i < 3; i++) w[i] = 0;
    for(int i = 0; i < m; i++) Wts[i] = 0.000001;

    char jnk = 'a';
    ifstream Counter(filename);
    
    if(!Counter){
      cerr << "Input file is missing!!  Please check input file." << endl;
      exit(0);
    }
    
    while(!Counter.eof()){
      Counter >> tmpString;
      if(tmpString[0] == '>') n++; 
    }

    ifstream in(filename);
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> jnk;
    if(jnk == '>') {
      in >> InputData[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	InputData[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    InputData[counter - 1].number = counter;
    in.putback(jnk);
    }
    
    ofstream out("m-weights.tax");
    ofstream out2("2-weights.tax");

    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }

   if(m == 2){
     int flag1 = 0, flag2 = 0, INDEX[4];
     double hold = 0, s = 0, zero = 0;
     char tmp[127];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	  ofstream OUT("dummy.out");
	  
// 	  if(!OUT){
// 	    cerr << "Error on writing dummy.out." << endl;
// 	    exit(1);
// 	    }
	  
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
	  //  system("./fasta2phylip dummy.out out.mfa");
	  system("./fastDNAml < out.test > out.tax");
	  
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	    
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) 
	    IN >> tmpString;
	  
	  IN >> tmp;
	  w[0] = atof(tmp);
	  
	  // cout << tmp << endl;
	    
	  if(flag1 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[0]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[0]; flag2 = 1;}

	  if(w[0] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[0];
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;

	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[1]; flag2 = 1;}
	  
	  if(w[1] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}
	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 

	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 
	    
	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    out.precision(10);
	    if(hold != 0) 
	      for(int i = 0; i < m; i++) Wts[i] = hold /2;

	    out << i + 1 << " " << j + 1 << " " << w[0] << " " << Wts[0] << " " << Wts[1] << endl;
	    s += w[0];
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) Wts[i] = 0.00001;
	    flag1 = 0; flag2 = 0;
	}
      }
   }

    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4];
      double hold = 0, s = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) flags[i] = 0;
      for(int i = 0; i < m; i++) Wts[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	    ofstream OUT("dummy.out");
	    
// 	    if(!OUT){
// 	      cerr << "Error on writing dummy.out." << endl;
// 	      exit(1);
// 	    }
	    
// 	    zero = 0;

// 	    for(int i = 0; i < m; i++){
// 	      OUT << ">" << InputData[index[i]].name << endl;
// 	      OUT << InputData[index[i]].DNA << endl;
// 	    }

// 	    OUT << ">" << InputData[index[2]].name << endl;
// 	    OUT << InputData[index[2]].DNA << endl;

	    // system("./fasta2phylip dummy.out out.mfa");
	    system("./fastDNAml < out.test > out.tax");
	    
	    ifstream IN("out.tax");
	    if(!IN){
	      cerr << "Error on reading out.tax." << endl;
	      exit(1);
	    }
	    
	    while(tmpString != "Limits") {
	      IN >> tmpString; 
	    }
	    for(int p = 0; p < 8; p++) 
	      IN >> tmpString;


	    IN >> tmp;
	    w[0] = atof(tmp);

	    if(w[0] != 0){
	      int sums = 0;
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name)
		  if(flags[i] == 0){
		    Wts[i] = w[0];
		    flags[i] = 1;
		  }
		sums += flags[i];
	      }
	      if(sums == 0) hold = w[0];
	    }

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;
	    if((tmpString == "**")|| (tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0;

	    IN >> tmp;
	    w[1] = atof(tmp);

	    w[0] += w[1];

	    if(w[1] != 0){
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name){
		  if(flags[i] == 0){
		    Wts[i] = w[1];
		    flags[i] = 1;
		  }
		  flag = 1;
		}
		if(flag == 0) hold = w[1]; 
	      }
	    }

	    flag = 0;

	    for(int j2 = 0; j2 < 2 * m - 3; j2++){
	      for(int p = 0; p < 4; p++) 
		IN >> tmpString;
	      if((tmpString == "**")|| (tmpString == "*"))
		zero = 1;
	      if(zero == 1) for(int p = 0; p < 2; p++) {
		zero = 0;
		IN >> tmpString;
	    }
	      
	      else 
		IN >> tmpString;
	      zero = 0;
	  
	      IN >> tmp;
	      w[1] = atof(tmp);

	      w[0] += w[1];

	      if(w[1] != 0){
		for(int i = 0; i < m; i++){
		  if(tmpString == InputData[index[i]].name){
		    if(flags[i] == 0){
		      Wts[i] = w[1];
		      flags[i] = 1;
		    }
		    flag = 1;
		  }
		  if(flag == 0) hold = w[1];
		}		 
	      }
	      flag = 0;
	    }

	      Wts[2] = hold;

	    out << index[0] + 1 << " " << index[1] + 1 << " " << index[2] + 1 << " " << w[0] << " ";
	    for(int i = 0; i < m; i++){
	      if(Wts[i] != 0) out << Wts[i] << " ";
	      else out << "0.000001 ";
	    }
	    out << endl;
	    s += w[0];
	    flag = 0;
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) flags[i] = 0;
	    for(int i = 0; i < m; i++) Wts[i] = 0;
	    hold = 0;
	    
	  }
	}
      }

    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) Wts[i] = 0;
      WW = numOfWeight;

      int index[WW][m];

      ifstream INDEX;
      getList(n, m);

      INDEX.open("ListOfSubsets");
      if(!INDEX){
	cerr<<"Error for reading ListOfSubsets." << endl;
	exit(0);
      }
      for(int i = 0; i < numOfWeight; i++)
	for(int j = 0; j < m; j++) INDEX >> index[i][j];
      for(int i = 0; i < numOfWeight; i++){
	  WriteFafsaToPhylip(n, m, InputData, index[i]);
// 	ofstream OUT("dummy.out");
	
// 	if(!OUT){
// 	  cerr << "Error on writing dummy.out." << endl;
// 	  exit(1);
// 	}
// 	for(int j = 0; j < m; j++){	    
// 	  OUT << ">" << InputData[index[i][j]].name << endl;
// 	  OUT << InputData[index[i][j]].DNA << endl;
// 	}
// 	system("./fasta2phylip dummy.out out.mfa");
	system("./fastDNAml < out.test > out.tax");
	
	ifstream IN("out.tax");
	if(!IN){
	  cerr << "Error on reading out.tax." << endl;
	  exit(1);
	}
	
	while(tmpString != "Limits") {
	  IN >> tmpString; 
	}
	for(int p = 0; p < 8; p++) IN >> tmpString;
	
	IN >> tmp;
	w[0] = atof(tmp);
	
	//cout << tmp << endl;

	if(w[0] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[0];
	  }
	}
      
	for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	if((tmpString == "**")|| (tmpString == "*"))zero = 1;
	if(zero == 1) for(int p = 0; p < 2; p++) {
	  zero = 0;
	  IN >> tmpString;
	}
	
	else 
	  IN >> tmpString;
	zero = 0;

	IN >> tmp;
	w[1] = atof(tmp);
	//cout << tmp << endl;
	w[0] += w[1];	
	
	if(w[1] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[1];
	  }
	}
	
	for(int j2 = 0; j2 < 2 * m - 5; j2++){
	  
	  for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	  if((tmpString == "**")|| (tmpString == "*"))
	    zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0;
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  //cout << tmp << endl;

	  w[0] += w[1];	
	  
	  if(w[1] != 0){
	    for(int j = 0; j < m; j++){
	      if(tmpString == InputData[index[i][j]].name)
		Wts[j] = w[1];
	    }	//cout << tmp << endl;
	  }
	}
	
	for(int j = 0; j < m; j++)   
	  out << index[i][j] + 1 << " ";
	out << w[0] << " ";
	
	for(int i = 0; i < m; i++){
	  if(Wts[i] != 0) out << Wts[i] << " ";
	  else out << "0.000001 ";
	}
	out << endl;
	s += w[0];
	w[0] = 0;
	w[1] = 0;
	w[2] = 0;
	for(int i = 0; i < m; i++) Wts[i] = 0;
      }
      system ("rm ListOfSubsets");

    }

    /*************************************************/

    double numOfWeight2 = 0, w1 =0, w2 = 0, w3 = 0;
    numOfWeight2 = (n - 1) * n / 2;
    char chartmp[127];

    //    out2 << n << " " << numOfWeight2 << endl;
//     for(int i = 0; i < n; i++){
//       out2 << InputData[i].name << endl;
//     }
    
    for(int i = 0; i < n; i++){
      for(int j = i + 1; j < n; j++){
	  int INDEX[4];
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  WriteFafsaToPhylip(n, 2, InputData, INDEX);
	  system("./fastDNAml < out.test > out.tax");
	 
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	  
	  int zero = 0;
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) IN >> tmpString;

	  IN >> chartmp;
	  w1 = atof(chartmp);
	  //cout << chartmp << endl;
	 
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 

	  IN >> chartmp;
	  w2  = atof(chartmp);
	  w1 += w2;
	  // cout << chartmp << endl;

	  //	  if(w1 == 0) for(int p = 0; p < 5; p++) IN >> tmpString;
	  //  else for(int p = 0; p < 6; p++) IN >> tmpString;

	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 
	    

	  IN >> chartmp;
	  w2 = atof(chartmp);
	  w1 += w2;
	  //cout << chartmp << endl;
	 
	  // if(w2 == 0) for(int p = 0; p < 5; p++) IN >> tmpString;
	  //else for(int p = 0; p < 6; p++) IN >> tmpString;

	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	    }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> chartmp;
	  w2 = atof(chartmp);
	  w1 += w2;
	  // cout << chartmp << endl;

	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	      IN >> tmpString;
	  zero = 0; 
	  
	  IN >> chartmp;
	  w2 = atof(chartmp);
	  w1 += w2;
	  // cout << chartmp << endl;

	  // out2 << i + 1 << " " << j + 1 << " " << w1 << endl;
	  out2 << w1 << " ";
	  w1 = 0;
	  w2 = 0;
	  w3 = 0;
      }
      
    }
      
    
    // system ("rm out.mfa");
    system ("rm checkpoint.*");
    system ("rm treefile.*");
}

/***************************************************************/

void ReadTaxa(Taxa * Data, int n, int m, int numOfWeights, ifstream & in, char ** names){
  char tmp[127];

  for(int i = 0; i < n; i++){
    in >> tmp;
    names[i] = new char[strlen(tmp) + 1];
    strcpy(names[i], tmp);
  }
  
  if(n <  m + 2){
    cerr << "The number of leaves must be more than m + 2. " << endl;
    exit(0);
  }
 
  for(int f = 0; f < numOfWeights; f++){
    for(int i = 0; i < m; i++) in >> Data[f].array[i];
    in >> Data[f].wt;// cout << Data[f].wt << endl;
    for(int i = 0; i < m; i++) in >> Data[f].edgeWts[i];
  }

//   cout << endl;
//   for(int j = 0; j < n; j++) cout << names[j] << endl;
//   for(int f = 0; f < numOfWeights; f++){
//     for(int i = 0; i < m; i++) cout << Data[f].array[i] << " ";
//     cout << Data[f].wt << " ";
//     for(int i = 0; i < m; i++) cout << Data[f].edgeWts[i] << " ";
//     cout << endl;
//   }

} 

/********************* Reading phylip file **********************************/

void ReadPhylFile(int & n, int & m, char* filename){
  int counter = 0;
    string tmpString;
    char name[127];
    double w[3], Wts[m];
    for(int i = 0; i < 3; i++) w[i] = 0;
    for(int i = 0; i < m; i++) Wts[i] = 0.000001;

    ifstream in(filename);

    if(!in){
      cerr << "Input file is missing for reading phylip format!!  Please check input file." << endl;
      exit(0);
    }
    
    cout << "Reading the input file...";
  
    in >> n >> counter;
    
    Alinements InputData[n];
    for(int i = 0; i < n; i++){
      in >> name;
      strcpy(InputData[i].name, name);

      in >> InputData[i].DNA; 
      //  cout << InputData[i].name << "     " <<  InputData[i].DNA << endl;
    }

    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << InputData[i].name << endl;
    }

   if(m == 2){
     int flag1 = 0, flag2 = 0, INDEX[4];
     double hold = 0, s = 0, zero = 0;
     char tmp[127];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	  ofstream OUT("dummy.out");
	  
// 	  if(!OUT){
// 	    cerr << "Error on writing dummy.out." << endl;
// 	    exit(1);
// 	    }
	  
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
// 	  OUT << ">" << InputData[i].name << endl;
// 	  OUT << InputData[i].DNA << endl;
// 	  OUT << ">" << InputData[j].name << endl;
// 	  OUT << InputData[j].DNA << endl;
	  //  system("./fasta2phylip dummy.out out.mfa");
	  system("./fastDNAml < out.test > out.tax");
	  
	  ifstream IN("out.tax");
	  if(!IN){
	    cerr << "Error on reading out.tax." << endl;
	    exit(1);
	  }
	    
	  while(tmpString != "Limits") {
	    IN >> tmpString; 
	  }
	  for(int p = 0; p < 8; p++) 
	    IN >> tmpString;
	  
	  IN >> tmp;
	  w[0] = atof(tmp);
	  
	  // cout << tmp << endl;
	    
	  if(flag1 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[0]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[0] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[0]; flag2 = 1;}

	  if(w[0] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[0];
	  
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;

	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[j].name){
		Wts[1] = w[1]; flag2 = 1;}
	  
	  if(w[1] != 0)
	    if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	  for(int p = 0; p < 4; p++) 
	    IN >> tmpString;
	  
	  if((tmpString == "**")||(tmpString == "*")) zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0; 
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  
	  //cout << tmp << endl;
	  
	  w[0] += w[1];
	  
	  if(flag1 == 0)
	    if(w[1] != 0)
	      if(tmpString == InputData[i].name){
		Wts[0] = w[1]; flag1 = 1;}
	  
	  if(flag2 == 0)
	    if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}
	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];
	    
	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 

	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;

	    if((tmpString == "**")||(tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0; 
	    
	    IN >> tmp;
	    w[1] = atof(tmp);

	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(flag1 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[i].name){
		  Wts[0] = w[1]; flag1 = 1;}
	    
	    if(flag2 == 0)
	      if(w[1] != 0)
		if(tmpString == InputData[j].name){
		  Wts[1] = w[1]; flag2 = 1;}

	    if(w[1] != 0)
	      if((flag1 == 0) && (flag2 == 0)) hold = w[1];

	    out.precision(10);
	    if(hold != 0) 
	      for(int i = 0; i < m; i++) Wts[i] = hold /2;

	    out << i + 1 << " " << j + 1 << " " << w[0] << " " << Wts[0] << " " << Wts[1] << endl;
	    s += w[0];
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) Wts[i] = 0.00001;
	    flag1 = 0; flag2 = 0;
	}
      }
   }

    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4];
      double hold = 0, s = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) flags[i] = 0;
      for(int i = 0; i < m; i++) Wts[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  WriteFafsaToPhylip(n, m, InputData, INDEX);
// 	    ofstream OUT("dummy.out");
	    
// 	    if(!OUT){
// 	      cerr << "Error on writing dummy.out." << endl;
// 	      exit(1);
// 	    }
	    
// 	    zero = 0;

// 	    for(int i = 0; i < m; i++){
// 	      OUT << ">" << InputData[index[i]].name << endl;
// 	      OUT << InputData[index[i]].DNA << endl;
// 	    }

// 	    OUT << ">" << InputData[index[2]].name << endl;
// 	    OUT << InputData[index[2]].DNA << endl;

	    // system("./fasta2phylip dummy.out out.mfa");
	    system("./fastDNAml < out.test > out.tax");
	    
	    ifstream IN("out.tax");
	    if(!IN){
	      cerr << "Error on reading out.tax." << endl;
	      exit(1);
	    }
	    
	    while(tmpString != "Limits") {
	      IN >> tmpString; 
	    }
	    for(int p = 0; p < 8; p++) 
	      IN >> tmpString;


	    IN >> tmp;
	    w[0] = atof(tmp);
	    // cout << tmp << endl;

	    if(w[0] != 0){
	      int sums = 0;
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name)
		  if(flags[i] == 0){
		    Wts[i] = w[0];
		    flags[i] = 1;
		  }
		sums += flags[i];
	      }
	      if(sums == 0) hold = w[0];
	    }

	    for(int p = 0; p < 4; p++) 
	      IN >> tmpString;
	    if((tmpString == "**")|| (tmpString == "*")) zero = 1;
	    if(zero == 1) for(int p = 0; p < 2; p++) {
	      zero = 0;
	      IN >> tmpString;
	    }

	    else 
	      IN >> tmpString;
	    zero = 0;

	    IN >> tmp;
	    w[1] = atof(tmp);
	    //cout << tmp << endl;

	    w[0] += w[1];

	    if(w[1] != 0){
	      for(int i = 0; i < m; i++){
		if(tmpString == InputData[index[i]].name){
		  if(flags[i] == 0){
		    Wts[i] = w[1];
		    flags[i] = 1;
		  }
		  flag = 1;
		}
		if(flag == 0) hold = w[1]; 
	      }
	    }

	    flag = 0;

	    for(int j2 = 0; j2 < 2 * m - 3; j2++){
	      for(int p = 0; p < 4; p++) 
		IN >> tmpString;
	      if((tmpString == "**")|| (tmpString == "*"))
		zero = 1;
	      if(zero == 1) for(int p = 0; p < 2; p++) {
		zero = 0;
		IN >> tmpString;
	    }
	      
	      else 
		IN >> tmpString;
	      zero = 0;
	  
	      IN >> tmp;
	      w[1] = atof(tmp);
	      //cout << tmp << endl;

	      w[0] += w[1];

	      if(w[1] != 0){
		for(int i = 0; i < m; i++){
		  if(tmpString == InputData[index[i]].name){
		    if(flags[i] == 0){
		      Wts[i] = w[1];
		      flags[i] = 1;
		    }
		    flag = 1;
		  }
		  if(flag == 0) hold = w[1];
		}		 
	      }
	      flag = 0;
	    }

	      Wts[2] = hold;

	    out << index[0] + 1 << " " << index[1] + 1 << " " << index[2] + 1 << " " << w[0] << " ";
	    for(int i = 0; i < m; i++){
	      if(Wts[i] != 0) out << Wts[i] << " ";
	      else out << "0.000001 ";
	    }
	    out << endl;
	    s += w[0];
	    flag = 0;
	    w[0] = 0;
	    w[1] = 0;
	    w[2] = 0;

	    for(int i = 0; i < m; i++) flags[i] = 0;
	    for(int i = 0; i < m; i++) Wts[i] = 0;
	    hold = 0;
	    
	  }
	}
      }

    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127];

      for(int i = 0; i < m; i++) Wts[i] = 0;
      WW = numOfWeight;

      int index[WW][m];

      ifstream INDEX;
      getList(n, m);

      INDEX.open("ListOfSubsets");
      if(!INDEX){
	cerr<<"Error for reading ListOfSubsets." << endl;
	exit(0);
      }
      for(int i = 0; i < numOfWeight; i++)
	for(int j = 0; j < m; j++) INDEX >> index[i][j];
      for(int i = 0; i < numOfWeight; i++){
	  WriteFafsaToPhylip(n, m, InputData, index[i]);
// 	ofstream OUT("dummy.out");
	
// 	if(!OUT){
// 	  cerr << "Error on writing dummy.out." << endl;
// 	  exit(1);
// 	}
// 	for(int j = 0; j < m; j++){	    
// 	  OUT << ">" << InputData[index[i][j]].name << endl;
// 	  OUT << InputData[index[i][j]].DNA << endl;
// 	}
// 	system("./fasta2phylip dummy.out out.mfa");
	system("./fastDNAml < out.test > out.tax");
	
	ifstream IN("out.tax");
	if(!IN){
	  cerr << "Error on reading out.tax." << endl;
	  exit(1);
	}
	
	while(tmpString != "Limits") {
	  IN >> tmpString; 
	}
	for(int p = 0; p < 8; p++) IN >> tmpString;
	
	IN >> tmp;
	w[0] = atof(tmp);
	
	//cout << tmp << endl;

	if(w[0] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[0];
	  }
	}
      
	for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	if((tmpString == "**")|| (tmpString == "*"))zero = 1;
	if(zero == 1) for(int p = 0; p < 2; p++) {
	  zero = 0;
	  IN >> tmpString;
	}
	
	else 
	  IN >> tmpString;
	zero = 0;

	IN >> tmp;
	w[1] = atof(tmp);
	//cout << tmp << endl;
	w[0] += w[1];	
	
	if(w[1] != 0){
	  for(int j = 0; j < m; j++){
	    if(tmpString == InputData[index[i][j]].name)
	      Wts[j] = w[1];
	  }
	}
	
	for(int j2 = 0; j2 < 2 * m - 5; j2++){
	  
	  for(int p = 0; p < 4; p++) 
	  IN >> tmpString;
	  if((tmpString == "**")|| (tmpString == "*"))
	    zero = 1;
	  if(zero == 1) for(int p = 0; p < 2; p++) {
	    zero = 0;
	    IN >> tmpString;
	  }
	  
	  else 
	    IN >> tmpString;
	  zero = 0;
	  
	  IN >> tmp;
	  w[1] = atof(tmp);
	  //cout << tmp << endl;

	  w[0] += w[1];	
	  
	  if(w[1] != 0){
	    for(int j = 0; j < m; j++){
	      if(tmpString == InputData[index[i][j]].name)
		Wts[j] = w[1];
	    }	//cout << tmp << endl;
	  }
	}
	
	for(int j = 0; j < m; j++)   
	  out << index[i][j] + 1 << " ";
	out << w[0] << " ";
	
	for(int i = 0; i < m; i++){
	  if(Wts[i] != 0) out << Wts[i] << " ";
	  else out << "0.000001 ";
	}
	out << endl;
	s += w[0];
	w[0] = 0;
	w[1] = 0;
	w[2] = 0;
	for(int i = 0; i < m; i++) Wts[i] = 0;
      }
      system ("rm ListOfSubsets");
      //  cout << endl << s << endl;
    }
    
    // system ("rm out.mfa");
    system ("rm checkpoint.*");
    system ("rm treefile.*");
}


/* This is for EMGNJ.... */

void ReadFMAFileNJforEMGNJ(int & n, int & m, char* filename, 
			   Alinements ** InputData){
   int counter = 0, num = 0;
   string tmpString;
   
   char jnk = 'a';
   ifstream Counter(filename);
   
   if(!Counter){
     cerr << "Input file is missing!!  Please check input file." << endl;
     exit(0);
   }
   
   // cout << "Reading the input file...";
   while(!Counter.eof()){
     Counter >> tmpString;
     if(tmpString[0] == '>') n++; 
   }
   
   if(n < m+1){
     cerr << "Error: Please check that m < n where n is the number of taxa." << endl << endl;
     exit(0);
   }
   
   
   ifstream in(filename);
   
   *InputData = new Alinements[n];
   for(int i = 0; i < n; i++){
     in >> jnk; 
     if(jnk == '>') {
       in >> (*InputData)[counter].name; 
     }
     jnk = 'a';
     while(jnk != '>'){
       if(jnk != 'a')     in.putback(jnk);
       if(in.eof()) jnk = '>';
       in >> tmpString; 
       if(jnk != '>')
	 (*InputData)[counter].DNA.append(tmpString);
       in >> jnk;
     }
     counter++;
     (*InputData)[counter - 1].number = counter;
     in.putback(jnk);
   }
}

void computeIniTreeforEMGNJ(int n, int m, char* freq, 
			   char* TsTv, double &tfreq, double &cfreq, 
			   double &afreq, double &gfreq, double trans,
			   Alinements ** InputData){
   int counter = 0, num = 0;
   string tmpString;
   double w[3];
    for(int i = 0; i < 3; i++) w[i] = 0;


    ofstream out("m-weights.tax");
    out.precision(6);

    if(!out){
      cerr << "Error on writing m-weights.tax." << endl;
      exit(1);
    }
    long int numOfWeight = 1;
    if(m <= n/2){
      for(int i = 0; i < m; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= m; i++) numOfWeight /= i;
    }
    else{
      int mm = n - m;
      for(int i = 0; i < mm; i++)  numOfWeight *= (n - i);
      for(int i = 1; i <= mm; i++) numOfWeight /= i;
    }

    out << n << " " << numOfWeight << endl;
    for(int i = 0; i < n; i++){
      out << (*InputData)[i].name << endl;
    }

   if(m == 2){
     int INDEX[4], length = 0;
     double s = 0, zero = 0;
     char tmp[127], *dna[4];
     for(int i = 0; i < 4; i++) 
       dna[i] = new char[(*InputData)[0].DNA.length() + 1];

      for(int i = 0; i < n; i++){
	for(int j = i + 1; j < n; j++){
	  INDEX[0] = i;
	  INDEX[1] = j;
	  INDEX[2] = i;
	  INDEX[3] = j;
	  StringToChar(n, 4, *InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	}
      } out << endl;
     for(int i = 0; i < 4; i++) 
       delete [] dna[i];
   }

  
    if(m == 3){
      int flags[m], index[m], flag = 0, zero = 0, INDEX[4], length = 0;
      double s = 0;
      char tmp[127], *dna[4];
      for(int i = 0; i < 4; i++) 
	dna[i] = new char[(*InputData)[0].DNA.length() + 1];

      for(int i = 0; i < m; i++) flags[i] = 0;

      for(index[0] = 0; index[0] < n; index[0]++){
	for(index[1] = index[0] + 1; index[1] < n; index[1]++){
	  for(index[2] = index[1] + 1; index[2] < n; index[2]++){
	    for(int k = 0; k < m; k++)
	      INDEX[k] = index[k];
	  INDEX[3] = index[m - 1];

	  StringToChar(n, 4, *InputData, INDEX, dna, length);

	  out.precision(10);
	  
	  out << fixed << getWts(4, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna) << " ";
	  num++;
	  //	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
	  }
	}
      }
      out << endl;
      for(int i = 0; i < 4; i++) 
	delete [] dna[i];
    }
     
    if(m >= 4) {
      int WW;
      double s = 0, zero = 0;
      char tmp[127], *dna[m];
      for(int i = 0; i < m; i++) 
	dna[i] = new char[(*InputData)[0].DNA.length() + 1];
      
      WW = numOfWeight;

      int length = 0;// *index[WW];
//       for(int i = 0; i < numOfWeight; i++) 
// 	index[i] = new int[m];

//      getList3(n, m, index);
      getList4(n, m, freq, TsTv, tfreq, cfreq, 
	 afreq, gfreq, trans, *InputData, dna, length,
	 num, numOfWeight, out);

//       for(int i = 0; i < numOfWeight; i++){
// 	StringToChar(n, m, InputData, index[i], dna, length);
// 	if(m != n){
// 	  out << fixed << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< " ";
// 	  num++;
// 	  cout << "\rComputing subtree weight: " << num << "/" << numOfWeight << "." << flush;
// 	}
// 	else {
// 	  //	  ofstream total("totalWeight");
// 	  // total << getWts(m, length, freq, TsTv, tfreq, cfreq, afreq, gfreq, trans, dna)<< endl;
// 	  exit(0);
// 	}

// 	delete [] index[i];
//       }
      out << endl;
      for(int i = 0; i < m; i++) 
	delete [] dna[i];
    }
    
}

