/*********************************************************************
 Author:  Ruriko Yoshida
 Date:    March 14th, 2006
 Update:  March 15th, 2006
 Program: This is software for compressing site patterns.

 Input:   The alignment formatted as the multi FASTA format form.

 Output:  List of site patterns.
 
 Modified for EMGNJ.

***********************************************************************/

#include <fstream.h>
#include <iostream.h>
#include <cstdlib>
#include <cstring>
#include <math.h>
#include <algorithm>
#include <time.h>

#include "cmpressed.h"
#include "Reading.h"
#include "Seq.h"


using namespace std;

void BST::clear(SiteNode *p) {
    if (p != 0) {
         clear(p->left);
         clear(p->right);
	 // p->node.clear();
         delete p;
     }
}

void BST::insert(const SitePattern& el) {
    SiteNode *p = root, *prev = 0;
    while (p != 0) {        // find a place for inserting new node;
        prev = p;
	if((p->node.Equal(el.site)) == 0){
	   break;
	}
        if ((p->node.Less(el.site)) == 0)
             p = p->right;
        else  p = p->left;
    }
    if (root == 0){    // tree is empty;
         root = new SiteNode(el);
    }
    else if ((prev->node.Equal(el.site)) == 0) {
      prev->node.num++;
/*       cout << "=" << endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << prev->node.site[j] <<" "; */
/*       cout <<  endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << el.site[j] <<" "; */
/*       cout <<  endl; */
    }
    else if ((prev->node.Less(el.site)) == 0){
      prev->right = new SiteNode(el);
      prev->right->parent = prev;
/*       cout << "<" << endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << prev->node.site[j] <<" "; */
/*       cout <<  endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << el.site[j] <<" "; */
/*       cout <<  endl; */
    }
    else if ((prev->node.Greater(el.site)) == 0){
      prev->left  = new SiteNode(el);
      prev->left->parent = prev;
/*       cout << ">" << endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << prev->node.site[j] <<" "; */
/*       cout <<  endl; */
/*       for(int j = 0; j < el.numOfTaxa; j++) */
/* 	cout << el.site[j] <<" "; */
/*       cout <<  endl; */
    }
    else if (prev->parent != 0)
      if((prev->parent->node.Equal(el.site)) == 0) {
	prev->node.num++;
      }

    else cout << "Error." << endl;
}


SitePattern* BST::search(SiteNode* p, const SitePattern& el) const {
    while (p != 0)
        if (p->node == el)
             return &p->node;
        else if (p->node > el)
             p = p->left;
        else p = p->right;
    return 0;
}

void BST::inorder(SiteNode *p) {
     if (p != 0) {
         inorder(p->left);
         visit(p);
         inorder(p->right);
     }
}

void BST::preorder(SiteNode *p) {
    if (p != 0) {
        visit(p);
        preorder(p->left);
        preorder(p->right);
    }
}

void BST::postorder(SiteNode* p) {
    if (p != 0) {
        postorder(p->left);
        postorder(p->right);
        visit(p);
    }
}

void BST::sum(SiteNode* p, int & total) {
    if (p != 0) {
        sum(p->left, total);
        sum(p->right, total);
        total += p->node.num;
    }
}

void BST::numOfNodes(SiteNode* p, int & total) {
    if (p != 0) {
        numOfNodes(p->left, total);
        numOfNodes(p->right, total);
        total++;
    }
}

void BST::setMatrix(SiteNode* p, int** total, int & index) {
    if (p != 0) {
        setMatrix(p->left, total, index);
        setMatrix(p->right, total, index);
	for(int i = 0; i < p->node.numOfTaxa; i++)
	  total[i][index] = p->node.site[i];
	total[p->node.numOfTaxa][index] =  p->node.num;
	index++;
    }
}

void ReadFMA(int & n, char* filename, Alinements ** InputData){
  ifstream in(filename);
  int counter = 0, num = 0;
  string tmpString;   char jnk = 'a';
  ifstream Counter(filename);
  
  if(!Counter){
    cerr << "Input file is missing!!  Please check input file." << endl;
    exit(0);
  }
  
  // cout << "Reading the input file...";
  while(!Counter.eof()){
    Counter >> tmpString;
    if(tmpString[0] == '>') n++; 
  }
  

  *InputData = new Alinements[n];
  for(int i = 0; i < n; i++){
    in >> jnk; 
    if(jnk == '>') {
      in >> (*InputData)[counter].name; 
    }
    jnk = 'a';
    while(jnk != '>'){
      if(jnk != 'a')     in.putback(jnk);
      if(in.eof()) jnk = '>';
      in >> tmpString; 
      if(jnk != '>')
	(*InputData)[counter].DNA.append(tmpString);
      in >> jnk;
    }
    counter++;
    (*InputData)[counter - 1].number = counter;
    in.putback(jnk);
  }
   
}

 
/* ----------------------------------------------------------------- */
int** cmpressed(int** Mat, int n, int & newLn, int length) {
  int **cmpMatrix;
 
  BST tree;
 
  SitePattern r;
  r.Allocate(n);
  for(int i = 0; i < n; i++)
    r.site[i] = Mat[i][0];
  tree.insert(r);  
  for(int i = 1; i < length; i++){
    SitePattern r2;
    r2.Allocate(n);
    for(int j = 0; j < n; j++) r2.site[j] = Mat[j][i];

    tree.insert(r2);

  }
  newLn = tree.numOfNodes(); 
  //cout << newLn << endl;
  cmpMatrix = new int*[n + 1];
  for(int i = 0; i < (n + 1); i++)
    cmpMatrix[i] = new int[newLn];
  tree.setMatrix(cmpMatrix);
//   for(int i = 0; i < (n + 1); i++){
//     for(int j = 0; j < newLn; j++)
//       cout << cmpMatrix[i][j] << " ";
//     cout << endl;
//   }

  return cmpMatrix;
}
